"""Argument data models."""
# pylint: disable=no-self-argument
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

from pydantic import DirectoryPath, Extra, Field, FilePath, validator

from .....config.models.utils import resolve_path_field
from .....utils import BaseModel
from ...base import HookArgsBaseModel

if TYPE_CHECKING:
    from typing import Callable


class DockerOptions(BaseModel):
    """Docker options."""

    disabled: bool = False
    """Explicitly disable the use of docker (default ``False``).

    If not disabled and Docker is unreachable, the hook will result in an error.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          docker:
            disabled: true

    """

    extra_files: List[str] = []
    """List of absolute file paths within the Docker container to copy into the
    deployment package.

    Some Python packages require extra OS libraries (``*.so``) files at runtime.
    These files need to be included in the deployment package for the Lambda Function to run.
    List the files here and the hook will handle copying them into the deployment package.

    The file name may end in a wildcard (``*``) to accommodate ``.so`` files that
    end in an variable number (see example below).

    If the file does not exist, it will result in an error.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          docker:
            extra_files:
              - /usr/lib64/mysql/libmysqlclient.so.*
              - /usr/lib64/libxmlsec1-openssl.so

    """

    file: Optional[FilePath] = None
    """Dockerfile to use to build an image for use in this process.

    This, ``image`` , or ``runtime`` must be provided.
    If not provided, ``image`` will be used.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          docker:
            file: Dockerfile

    """

    image: Optional[str] = None
    """Docker image to use. If the image does not exist locally, it will be pulled.

    This, ``file`` (takes precedence), or ``runtime`` must be provided.
    If only ``runtime`` is provided, it will be used to determine the appropriate
    image to use.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          docker:
            image: public.ecr.aws/sam/build-python3.9:latest

    """

    name: Optional[str] = None
    """When providing a Dockerfile, this will be the name applied to the resulting image.
    It is the equivalent to ``name`` in the ``name:tag`` syntax of the
    ``docker build [--tag, -t]`` command option.

    If not provided, a default image name is used.

    This field is ignore unless ``file`` is provided.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          docker:
            file: Dockerfile
            name: ${namespace}.runway.awslambda

    """

    pull: bool = True
    """Always download updates to the specified image before use.
    When building an image, the ``FROM`` image will be updated during the build
    process  (default ``True``).

    .. rubric:: Example
    .. code-block:: yaml

        args:
          docker:
            pull: false

    """

    class Config:
        """Model configuration."""

        extra = Extra.ignore

    _resolve_path_fields = cast(
        "classmethod[Callable[..., Any]]",
        validator("file", allow_reuse=True)(resolve_path_field),
    )


class AwsLambdaHookArgs(HookArgsBaseModel):
    """Base class for AWS Lambda hook arguments."""

    bucket_name: str
    """Name of the S3 Bucket where deployment package is/will  be stored.
    The Bucket must be in the same region the Lambda Function is being deployed in.

    """

    cache_dir: Optional[Path] = None
    """Explicitly define the directory location.
    Must be an absolute path or it will be relative to the CFNgin module directory.

    """

    compatible_architectures: Optional[List[str]] = None
    """A list of compatible instruction set architectures.
    (https://docs.aws.amazon.com/lambda/latest/dg/foundation-arch.html)

    Only used by Lambda Layers.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          compatible_architectures:
            - x86_64
            - arm64

    """

    compatible_runtimes: Optional[List[str]] = None
    """A list of compatible function runtimes.
    When provided, the ``runtime`` being used to build the deployment package
    must be included in the list or an error will be raised.

    Only used by Lambda Layers.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          compatible_runtimes:
            - python3.8
            - python3.9

    """

    docker: DockerOptions = DockerOptions()
    """Docker options."""

    extend_gitignore: List[str] = []
    """gitignore rules that should be added to the rules already defined in a
    ``.gitignore`` file in the source code directory.
    This can be used with or without an existing file.
    Files that match a gitignore rule will not be included in the deployment package.

    ``.git/`` & ``.gitignore`` will always be added.

    .. important:: This only applies to files in the ``source_code`` directory.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          extend_gitignore:
            - cfngin.yml
            - poetry.lock
            - poetry.toml
            - pyproject.toml

    """

    license: Optional[str] = Field(default=None, max_length=256)
    """The layer's software license. Can be any of the following:

    - A SPDX license identifier (e.g. ``Apache-2.0``).
    - The URL of a license hosted on the internet (e.g.
      ``https://opensource.org/licenses/Apache-2.0``).
    - The full text of the license.

    Only used by Lambda Layers.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          license: Apache-2.0

    """

    object_prefix: Optional[str] = None
    """Prefix to add to the S3 Object key.

    The object will always be prefixed with ``awslambda/functions``.
    If provided, the value will be added to the end of the static prefix
    (e.g. ``awslambda/<functions|layers>/<object_prefix>/<file name>``).

    """

    runtime: Optional[str] = None
    """Runtime of the Lambda Function
    (https://docs.aws.amazon.com/lambda/latest/dg/lambda-runtimes.html).

    This, ``docker.file``, or ``docker.image`` must be provided.
    If ``docker.disabled``, this field is required.

    When provided, the runtime available on the build system (Docker container
    or localhost) will be checked against this value. If they do not match,
    an error will be raised.

    If the defined or detected runtime ever changes so that it no longer
    matches the deployment package in S3, the deployment package in S3 will be
    deleted and a new one will be built and uploaded.

    """

    slim: bool = True
    """Automatically remove information and caches from dependencies (default ``True``).
    This is done by applying gitignore rules to the dependencies.
    These rules vary by language/runtime.

    """

    source_code: DirectoryPath
    """Path to the Lambda Function source code.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          source_code: ./my/package

    """

    use_cache: bool = True
    """Whether to use a cache directory with pip that will persist builds (default ``True``)."""

    _resolve_path_fields = cast(
        "classmethod[Callable[..., Any]]",
        validator("cache_dir", "source_code", allow_reuse=True)(resolve_path_field),
    )

    @validator("runtime", always=True, allow_reuse=True)
    def _validate_runtime_or_docker(
        cls, v: Optional[str], values: Dict[str, Any]
    ) -> Optional[str]:
        """Validate that either runtime is provided or Docker image is provided."""
        if v:  # if runtime was provided, we don't need to check anything else
            return v
        docker: DockerOptions = values["docker"]
        if docker.disabled:
            raise ValueError("runtime must be provided if docker.disabled is True")
        if not (docker.file or docker.image):
            raise ValueError("docker.file, docker.image, or runtime is required")
        return v


class PythonHookArgs(AwsLambdaHookArgs):
    """Hook arguments for a Python AWS Lambda deployment package."""

    extend_pip_args: Optional[List[str]] = None
    """Additional arguments that should be passed to ``pip install``.

    .. important::
      When providing this field, be careful not to duplicate any of the arguments
      passed by this hook (e.g. ``--requirements``, ``--target``, ``--no-input``).
      Providing duplicate arguments will result in an error.

    .. rubric:: Example
    .. code-block:: yaml

        args:
          extend_pip_args:
            - '--proxy'
            - '[user:passwd@]proxy.server:port'

    """

    slim: bool = True
    """Automatically remove information and caches from dependencies (default ``True``).
    This is done by applying the following gitignore rules to the dependencies:

    - ``**/*.dist-info*``
    - ``**/*.py[c|d|i|o]``
    - ``**/*.so``
    - ``**/__pycache__*``

    """

    strip: bool = True
    """Whether or not to strip binary files from the dependencies (default ``True``).
    This only takes effect if ``slim: true``.

    If false, the gitignore rule ``**/*.so`` is not used.

    """

    use_pipenv: bool = True
    """Whether pipenv should be used if determined appropriate."""

    use_poetry: bool = True
    """Whether poetry should be used if determined appropriate."""

    class Config:
        """Model configuration."""

        extra = Extra.ignore



"""Runway static site Module parameters."""
# pylint: disable=no-self-argument
from __future__ import annotations

from typing import Dict, List, Optional, Union

from pydantic import Extra, Field, validator

from ....config.models.base import ConfigProperty


class RunwayStaticSiteCustomErrorResponseDataModel(ConfigProperty):
    """Model for Runway stat site Module staticsite_custom_error_responses parameter item.

    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-cloudfront-distribution-customerrorresponse.html

    """

    ErrorCachingMinTTL: Optional[int] = None
    ErrorCode: Optional[int] = None
    ResponseCode: Optional[int] = None
    ResponsePagePath: Optional[str] = None

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway static site Module staticsite_custom_error_responses parameter item."


class RunwayStaticSiteLambdaFunctionAssociationDataModel(ConfigProperty):
    """Model for Runway stat site Module staticsite_lambda_function_associations parameter item.

    Attributes:
        arn: Lambda function ARN.
        type: Association type.

    """

    arn: str
    type: str

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway static site Module staticsite_lambda_function_associations parameter item."  # noqa


class RunwayStaticSiteModuleParametersDataModel(ConfigProperty):
    """Model for Runway static site Module parameters.

    Attributes:
        acmcert_arn: The certificate arn used for any alias domains supplied.
            This is a requirement when supplying any custom domain.
        additional_redirect_domains: Additional domains (beyond the ``aliases``
            domains or the CloudFront URL if no aliases are provided) that will
            be authorized by the Auth@Edge UserPool AppClient.
        aliases: Any custom domains that should be added to the CloudFront
            Distribution.
        auth_at_edge: Auth@Edge make the static site private by placing it behind
            an authorization wall.
        cf_disable: Wether deployment of the CloudFront Distribution should be
            disabled.
        compress: Whether the CloudFront default cache behavior will automatically
            compress certain files.
        cookie_settings: The default cookie settings for retrieved tokens and
            generated nonce's.
        create_user_pool: Wether to create a User Pool for the Auth@Edge
            configuration.
        custom_error_responses: Define custom error responses.
        enable_cf_logging: Enable CloudFront logging.
        http_headers: Headers that should be sent with each origin response.
        lambda_function_associations: This allows the user to deploy custom
            Lambda@Edge associations with their pre-build function versions.
        namespace: The unique namespace for the deployment.
        non_spa: Wether this site is a single page application (SPA).
        oauth_scopes: Scope is a mechanism in OAuth 2.0 to limit an application's
            access to a user's account.
        redirect_path_auth_refresh: The path that a user is redirected to when
            their authorization tokens have expired (1 hour).
        redirect_path_sign_in: The path that a user is redirected to after sign-in.
        redirect_path_sign_out: The path that a user is redirected to after sign-out.
        required_group: Name of Cognito User Pool group of which users must be a
            member to be granted access to the site. If ``None``, allows all
            UserPool users to have access.
        rewrite_directory_index: Deploy a Lambda@Edge function designed to
            rewrite directory indexes.
        role_boundary_arn: Defines an IAM Managed Policy that will be set as the
            permissions boundary for any IAM Roles created to support the site.
        service_role: IAM role that CloudFormation will use.
        sign_out_url: The path a user should access to sign themselves out of the
            application.
        supported_identity_providers: A comma delimited list of the User Pool
            client identity providers.
        user_pool_arn: The ARN of a pre-existing Cognito User Pool to use with
            Auth@Edge.
        web_acl: The ARN of a web access control list (web ACL) to associate with
            the CloudFront Distribution.

    """

    acmcert_arn: Optional[str] = Field(default=None, alias="staticsite_acmcert_arn")
    additional_redirect_domains: List[str] = Field(
        default=[], alias="staticsite_additional_redirect_domains"
    )
    aliases: List[str] = Field(default=[], alias="staticsite_aliases")
    auth_at_edge: bool = Field(default=False, alias="staticsite_auth_at_edge")
    cf_disable: bool = Field(default=False, alias="staticsite_cf_disable")
    compress: bool = Field(default=True, alias="staticsite_compress")
    cookie_settings: Dict[str, str] = Field(
        default={
            "idToken": "Path=/; Secure; SameSite=Lax",
            "accessToken": "Path=/; Secure; SameSite=Lax",
            "refreshToken": "Path=/; Secure; SameSite=Lax",
            "nonce": "Path=/; Secure; HttpOnly; Max-Age=1800; SameSite=Lax",
        },
        alias="staticsite_cookie_settings",
    )
    create_user_pool: bool = Field(default=False, alias="staticsite_create_user_pool")
    custom_error_responses: List[RunwayStaticSiteCustomErrorResponseDataModel] = Field(
        default=[], alias="staticsite_custom_error_responses"
    )
    enable_cf_logging: bool = Field(default=True, alias="staticsite_enable_cf_logging")
    http_headers: Dict[str, str] = Field(
        default={
            "Content-Security-Policy": "default-src https: 'unsafe-eval' 'unsafe-inline'; "
            "font-src 'self' 'unsafe-inline' 'unsafe-eval' data: https:; "
            "object-src 'none'; "
            "connect-src 'self' https://*.amazonaws.com https://*.amazoncognito.com",
            "Strict-Transport-Security": "max-age=31536000; "
            "includeSubdomains; "
            "preload",
            "Referrer-Policy": "same-origin",
            "X-XSS-Protection": "1; mode=block",
            "X-Frame-Options": "DENY",
            "X-Content-Type-Options": "nosniff",
        },
        alias="staticsite_http_headers",
    )
    lambda_function_associations: List[
        RunwayStaticSiteLambdaFunctionAssociationDataModel
    ] = Field(default=[], alias="staticsite_lambda_function_associations")
    namespace: str
    non_spa: bool = Field(default=False, alias="staticsite_non_spa")
    oauth_scopes: List[str] = Field(
        default=[
            "phone",
            "email",
            "profile",
            "openid",
            "aws.cognito.signin.user.admin",
        ],
        alias="staticsite_oauth_scopes",
    )
    redirect_path_auth_refresh: str = Field(
        default="/refreshauth", alias="staticsite_redirect_path_auth_refresh"
    )
    redirect_path_sign_in: str = Field(
        default="/parseauth", alias="staticsite_redirect_path_sign_in"
    )
    redirect_path_sign_out: str = Field(
        default="/", alias="staticsite_redirect_path_sign_out"
    )
    required_group: Optional[str] = Field(
        default=None, alias="staticsite_required_group"
    )
    rewrite_directory_index: Optional[str] = Field(
        default=None, alias="staticsite_rewrite_directory_index"
    )
    role_boundary_arn: Optional[str] = Field(
        default=None, alias="staticsite_role_boundary_arn"
    )
    service_role: Optional[str] = Field(
        default=None, alias="cloudformation_service_role"
    )
    sign_out_url: str = Field(default="/signout", alias="staticsite_sign_out_url")
    supported_identity_providers: List[str] = Field(
        default=["COGNITO"], alias="staticsite_supported_identity_providers"
    )
    user_pool_arn: Optional[str] = Field(default=None, alias="staticsite_user_pool_arn")
    web_acl: Optional[str] = Field(default=None, alias="staticsite_web_acl")

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.ignore
        title = "Runway static site Module parameters."

    @validator(
        "additional_redirect_domains",
        "aliases",
        "supported_identity_providers",
        pre=True,
    )
    def _convert_comma_delimited_list(cls, v: Union[List[str], str]) -> List[str]:
        """Convert comma delimited lists to a string."""
        if isinstance(v, str):
            return [i.strip() for i in v.split(",")]
        return v



"""CFNgin config."""
from __future__ import annotations

import logging
import re
import sys
from pathlib import Path
from string import Template
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Union,
    cast,
)

import yaml

from ..cfngin import exceptions
from ..cfngin.lookups.registry import register_lookup_handler
from ..cfngin.utils import SourceProcessor
from ..exceptions import ConfigNotFound
from ..utils import merge_dicts
from .components.runway import (
    RunwayDeploymentDefinition,
    RunwayTestDefinition,
    RunwayVariablesDefinition,
)
from .models.cfngin import (
    CfnginConfigDefinitionModel,
    CfnginHookDefinitionModel,
    CfnginPackageSourcesDefinitionModel,
    CfnginStackDefinitionModel,
)
from .models.runway import RunwayConfigDefinitionModel, RunwayFutureDefinitionModel

if TYPE_CHECKING:
    from packaging.specifiers import SpecifierSet
    from pydantic import BaseModel

LOGGER = logging.getLogger(__name__)


class BaseConfig:
    """Base class for configurations."""

    file_path: Path
    _data: BaseModel

    def __init__(self, data: BaseModel, *, path: Optional[Path] = None) -> None:
        """Instantiate class.

        Args:
            data: The data model of the config file.
            path: Path to the config file.

        """
        self._data = data.copy()
        self.file_path = path.resolve() if path else Path.cwd()

    def dump(
        self,
        *,
        by_alias: bool = False,
        exclude: Optional[
            Union[AbstractSet[Union[int, str]], Mapping[Union[int, str], Any]]
        ] = None,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_unset: bool = True,
        include: Optional[
            Union[AbstractSet[Union[int, str]], Mapping[Union[int, str], Any]]
        ] = None,
    ) -> str:
        """Dump model to a YAML string.

        Args:
            by_alias: Whether field aliases should be used as keys in the
                returned dictionary.
            exclude: Fields to exclude from the returned dictionary.
            exclude_defaults: Whether fields which are equal to their default
                values (whether set or otherwise) should be excluded from
                the returned dictionary.
            exclude_none: Whether fields which are equal to None should be
                excluded from the returned dictionary.
            exclude_unset: Whether fields which were not explicitly set when
                creating the model should be excluded from the returned
                dictionary.
            include: Fields to include in the returned dictionary.

        """
        return yaml.dump(
            self._data.dict(
                by_alias=by_alias,
                exclude=exclude,  # type: ignore
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                exclude_unset=exclude_unset,
                include=include,  # type: ignore
            ),
            default_flow_style=False,
        )

    @classmethod
    def find_config_file(cls, path: Path) -> Optional[Path]:
        """Find a config file in the provided path.

        Args:
            path: The path to search for a config file.

        """
        raise NotImplementedError  # cov: ignore


class CfnginConfig(BaseConfig):
    """Python representation of a CFNgin config file.

    This is used internally by CFNgin to parse and validate a YAML formatted
    CFNgin configuration file, but can also be used in scripts to generate a
    CFNgin config file before handing it off to CFNgin to deploy/destroy.

    Example::

        from runway.cfngin.config import dump, Config, Stack

        vpc = Stack({
            "name": "vpc",
            "class_path": "blueprints.VPC"})

        config = Config()
        config.namespace = "prod"
        config.stacks = [vpc]

        print dump(config)

    """

    EXCLUDE_REGEX = r"runway(\..*)?\.(yml|yaml)"
    """Regex for file names to exclude when looking for config files."""

    EXCLUDE_LIST = ["bitbucket-pipelines.yml", "buildspec.yml", "docker-compose.yml"]
    """Explicit files names to ignore when looking for config files."""

    cfngin_bucket: Optional[str]
    """Bucket to use for CFNgin resources. (e.g. CloudFormation templates).
    May be an empty string.
    """

    cfngin_bucket_region: Optional[str]
    """Explicit region to use for :attr:`CfnginConfig.cfngin_bucket`"""

    cfngin_cache_dir: Path
    """Local directory to use for caching."""

    log_formats: Dict[str, str]
    """Custom formatting for log messages."""

    lookups: Dict[str, str]
    """Register custom lookups."""

    mappings: Dict[str, Dict[str, Dict[str, Any]]]
    """Mappings that will be added to all stacks."""

    namespace: str
    """Namespace to prepend to everything."""

    namespace_delimiter: str
    """Character used to separate :attr:`CfnginConfig.namespace` and anything it prepends."""

    package_sources: CfnginPackageSourcesDefinitionModel
    """Remote source locations."""

    persistent_graph_key: Optional[str] = None
    """S3 object key were the persistent graph is stored."""

    post_deploy: List[CfnginHookDefinitionModel]
    """Hooks to run after a deploy action."""

    post_destroy: List[CfnginHookDefinitionModel]
    """Hooks to run after a destroy action."""

    pre_deploy: List[CfnginHookDefinitionModel]
    """Hooks to run before a deploy action."""

    pre_destroy: List[CfnginHookDefinitionModel]
    """Hooks to run before a destroy action."""

    service_role: Optional[str]
    """IAM role for CloudFormation to use."""

    stacks: List[CfnginStackDefinitionModel]
    """Stacks to be processed."""

    sys_path: Optional[Path]
    """Relative or absolute path to use as the work directory."""

    tags: Optional[Dict[str, str]]
    """Tags to apply to all resources."""

    template_indent: int
    """Spaces to use per-indent level when outputting a template to json."""

    _data: CfnginConfigDefinitionModel

    def __init__(
        self,
        data: CfnginConfigDefinitionModel,
        *,
        path: Optional[Path] = None,
        work_dir: Optional[Path] = None,
    ) -> None:
        """Instantiate class.

        Args:
            data: The data model of the config file.
            path: Path to the config file.
            work_dir: Working directory.

        """
        super().__init__(data, path=path)

        self.cfngin_bucket = self._data.cfngin_bucket
        self.cfngin_bucket_region = self._data.cfngin_bucket_region
        if self._data.cfngin_cache_dir:
            self.cfngin_cache_dir = self._data.cfngin_cache_dir
        elif work_dir:
            self.cfngin_cache_dir = work_dir / "cache"
        elif path:
            self.cfngin_cache_dir = path.parent / ".runway" / "cache"
        else:
            self.cfngin_cache_dir = Path().cwd() / ".runway" / "cache"
        self.log_formats = self._data.log_formats
        self.lookups = self._data.lookups
        self.mappings = self._data.mappings
        self.namespace = self._data.namespace
        self.namespace_delimiter = self._data.namespace_delimiter
        self.package_sources = self._data.package_sources
        self.persistent_graph_key = self._data.persistent_graph_key
        self.post_deploy = cast(List[CfnginHookDefinitionModel], self._data.post_deploy)
        self.post_destroy = cast(
            List[CfnginHookDefinitionModel], self._data.post_destroy
        )
        self.pre_deploy = cast(List[CfnginHookDefinitionModel], self._data.pre_deploy)
        self.pre_destroy = cast(List[CfnginHookDefinitionModel], self._data.pre_destroy)
        self.service_role = self._data.service_role
        self.stacks = cast(List[CfnginStackDefinitionModel], self._data.stacks)
        self.sys_path = self._data.sys_path
        self.tags = self._data.tags
        self.template_indent = self._data.template_indent

    def load(self) -> None:
        """Load config options into the current environment/session."""
        if self.sys_path:
            LOGGER.debug("appending to sys.path: %s", self.sys_path)
            sys.path.append(str(self.sys_path))
            LOGGER.debug("sys.path: %s", sys.path)
        if self.lookups:
            for key, handler in self.lookups.items():
                register_lookup_handler(key, handler)

    @classmethod
    def find_config_file(  # type: ignore pylint: disable=arguments-differ
        cls, path: Optional[Path] = None, *, exclude: Optional[List[str]] = None
    ) -> List[Path]:
        """Find a config file in the provided path.

        Args:
            path: The path to search for a config file.
            exclude: List of file names to exclude. This list is appended to
                the global exclude list.

        Raises:
            ConfigNotFound: Could not find a config file in the provided path.
            ValueError: More than one config file found in the provided path.

        """
        if not path:
            path = Path.cwd()
        elif path.is_file():
            return [path]

        exclude = exclude or []
        result: List[Path] = []
        exclude.extend(cls.EXCLUDE_LIST)

        yml_files = list(path.glob("*.yml"))
        yml_files.extend(list(path.glob("*.yaml")))

        for f in yml_files:
            if (
                re.match(cls.EXCLUDE_REGEX, f.name)
                or f.name in exclude
                or f.name.startswith(".")
            ):
                continue  # cov: ignore
            result.append(f)
        result.sort()
        return result

    @classmethod
    def parse_file(
        cls,
        *,
        path: Optional[Path] = None,
        file_path: Optional[Path] = None,
        parameters: Optional[MutableMapping[str, Any]] = None,
        work_dir: Optional[Path] = None,
        **kwargs: Any,
    ) -> CfnginConfig:
        """Parse a YAML file to create a config object.

        Args:
            path: The path to search for a config file.
            file_path: Exact path to a file to parse.
            parameters: Values to use when resolving a raw config.
            work_dir: Explicit working directory.

        Raises:
            ConfigNotFound: Provided config file was not found.

        """
        if file_path:
            if not file_path.is_file():
                raise ConfigNotFound(path=file_path)
            return cls.parse_raw(
                file_path.read_text(),
                path=file_path,
                parameters=parameters or {},
                work_dir=work_dir,
                **kwargs,
            )
        if path:
            found = cls.find_config_file(path)
            if len(found) > 1:
                raise ValueError(f"more than one config files found: {found}")
            return cls.parse_file(
                file_path=found[0],
                parameters=parameters or {},
                work_dir=work_dir,
                **kwargs,
            )
        raise ValueError("must provide path or file_path")

    @classmethod
    def parse_obj(
        cls, obj: Any, *, path: Optional[Path] = None, work_dir: Optional[Path] = None
    ) -> CfnginConfig:
        """Parse a python object.

        Args:
            obj: A python object to parse as a CFNgin config.
            path: The path to the config file that was parsed into the object.
            work_dir: Working directory.

        """
        return cls(
            CfnginConfigDefinitionModel.parse_obj(obj), path=path, work_dir=work_dir
        )

    @classmethod
    def parse_raw(
        cls,
        data: str,
        *,
        parameters: Optional[MutableMapping[str, Any]] = None,
        path: Optional[Path] = None,
        skip_package_sources: bool = False,
        work_dir: Optional[Path] = None,
    ) -> CfnginConfig:
        """Parse raw data.

        Args:
            data: The raw data to parse.
            parameters: Values to use when resolving a raw config.
            path: The path to search for a config file.
            skip_package_sources: Skip processing package sources.
            work_dir: Explicit working directory.

        """
        if not parameters:
            parameters = {}
        pre_rendered = cls.resolve_raw_data(data, parameters=parameters)
        if skip_package_sources:
            return cls.parse_obj(yaml.safe_load(pre_rendered))
        config_dict = yaml.safe_load(
            cls.process_package_sources(
                pre_rendered, parameters=parameters, work_dir=work_dir
            )
        )
        return cls.parse_obj(config_dict, path=path)

    @classmethod
    def process_package_sources(
        cls,
        raw_data: str,
        *,
        parameters: Optional[MutableMapping[str, Any]] = None,
        work_dir: Optional[Path] = None,
    ) -> str:
        """Process the package sources defined in a rendered config.

        Args:
            raw_data: Raw configuration data.
            cache_dir: Directory to use when caching remote sources.
            parameters: Values to use when resolving a raw config.
            work_dir: Explicit working directory.

        """
        config = yaml.safe_load(raw_data) or {}
        processor = SourceProcessor(
            sources=CfnginPackageSourcesDefinitionModel.parse_obj(
                config.get("package_sources", {})  # type: ignore
            ),
            cache_dir=Path(
                config.get(
                    "cfngin_cache_dir", (work_dir or Path().cwd() / ".runway") / "cache"
                )
            ),
        )
        processor.get_package_sources()
        if processor.configs_to_merge:
            for i in processor.configs_to_merge:
                LOGGER.debug("merging in remote config: %s", i)
                with open(i, "rb") as opened_file:
                    config = merge_dicts(yaml.safe_load(opened_file), config)
            return cls.resolve_raw_data(yaml.dump(config), parameters=parameters or {})
        return raw_data

    @staticmethod
    def resolve_raw_data(
        raw_data: str, *, parameters: Optional[MutableMapping[str, Any]] = None
    ) -> str:
        """Resolve raw data.

        Args:
            raw_data: Raw configuration data.
            parameters: Values to use when resolving a raw config.

        Raises:
            MissingEnvironment: A value required by the config was not provided
                in parameters.

        """
        if not parameters:
            parameters = {}
        template = Template(raw_data)
        try:
            rendered = template.substitute(**parameters)
        except KeyError as err:
            raise exceptions.MissingEnvironment(err.args[0]) from None
        except ValueError:
            rendered = template.safe_substitute(**parameters)
        return rendered


class RunwayConfig(BaseConfig):
    """Python representation of a Runway config file."""

    ACCEPTED_NAMES = ["runway.yml", "runway.yaml"]

    deployments: List[RunwayDeploymentDefinition]
    file_path: Path
    future: RunwayFutureDefinitionModel
    ignore_git_branch: bool
    runway_version: Optional[SpecifierSet]
    tests: List[RunwayTestDefinition[Any]]
    variables: RunwayVariablesDefinition

    _data: RunwayConfigDefinitionModel

    def __init__(
        self, data: RunwayConfigDefinitionModel, *, path: Optional[Path] = None
    ) -> None:
        """Instantiate class.

        Args:
            data: The data model of the config file.
            path: Path to the config file.

        """
        super().__init__(data, path=path)
        self.deployments = [
            RunwayDeploymentDefinition(d) for d in self._data.deployments
        ]
        self.future = self._data.future
        self.ignore_git_branch = self._data.ignore_git_branch
        self.runway_version = self._data.runway_version
        self.tests = [RunwayTestDefinition(t) for t in self._data.tests]
        self.variables = RunwayVariablesDefinition(self._data.variables)

    @classmethod
    def find_config_file(cls, path: Path) -> Path:
        """Find a config file in the provided path.

        Args:
            path: The path to search for a config file.

        Raises:
            ConfigNotFound: Could not find a config file in the provided path.
            ValueError: More than one config file found in the provided path.

        """
        match = list(path.glob("runway.y*"))
        if not match or all(f.name not in cls.ACCEPTED_NAMES for f in match):
            match = list(path.parent.glob("runway.y*"))
        found = [f for f in match if f.is_file() and f.name in cls.ACCEPTED_NAMES]
        if not found:
            raise ConfigNotFound(looking_for=cls.ACCEPTED_NAMES, path=path)
        if len(found) != 1:
            raise ValueError(f"more than one config files found: {found}")
        return found[0]

    @classmethod
    def parse_file(
        cls,
        *,
        path: Optional[Path] = None,
        file_path: Optional[Path] = None,
        **kwargs: Any,
    ) -> RunwayConfig:
        """Parse a YAML file to create a config object.

        Args:
            path: The path to search for a config file.
            file_path: Exact path to a file to parse.

        Raises:
            ConfigNotFound: Provided config file was not found.
            ValueError: path and file_path were both excluded.

        """
        if file_path:
            if not file_path.is_file():
                raise ConfigNotFound(path=file_path)
            return cls.parse_obj(
                yaml.safe_load(file_path.read_text()), path=file_path, **kwargs
            )
        if path:
            return cls.parse_file(file_path=cls.find_config_file(path), **kwargs)
        raise ValueError("must provide path or file_path")

    @classmethod
    def parse_obj(cls, obj: Any, *, path: Optional[Path] = None) -> RunwayConfig:
        """Parse a python object into a config object.

        Args:
            obj: The object to be parsed.
            path: Path to the file the object was parsed from.

        """
        return cls(RunwayConfigDefinitionModel.parse_obj(obj), path=path)



"""Refresh authorization token for new credentials."""
import logging
import traceback
from urllib.parse import parse_qs

from shared import (  # noqa pylint: disable=import-error
    create_error_html,
    extract_and_parse_cookies,
    generate_cookie_headers,
    get_config,
    http_post_with_retry,
)

LOGGER = logging.getLogger(__name__)
CONFIG = get_config()


def handler(event, _context):
    """Handle the authorization refresh.

    Args:
        event: The Lambda Event.
        _context (Any): Lambda context object.

    """
    request = event["Records"][0]["cf"]["request"]
    domain_name = request["headers"]["host"][0]["value"]
    redirected_from_uri = f"https://{domain_name}"

    try:
        parsed_qs = parse_qs(request.get("querystring"))
        requested_uri = parsed_qs.get("requestedUri")[0]
        current_nonce = parsed_qs.get("nonce")[0]
        # Add the requested uri path to the main
        redirected_from_uri += requested_uri or ""

        cookies = extract_and_parse_cookies(request.get("headers"), CONFIG["client_id"])

        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        tokens = {
            "id_token": cookies["id_token"],
            "access_token": cookies["access_token"],
            "refresh_token": cookies["refresh_token"],
        }

        validate_refresh_request(current_nonce, cookies["nonce"], tokens)

        try:
            # Request new tokens based on the refresh_token
            body = {
                "grant_type": "refresh_token",
                "client_id": CONFIG["client_id"],
                "refresh_token": tokens.get("refresh_token"),
            }
            res = http_post_with_retry(
                f"https://{CONFIG['cognito_auth_domain']}/oauth2/token",
                body,
                headers,
            )
            tokens["id_token"] = res.get("id_token")
            tokens["access_token"] = res.get("access_token")
            cookie_headers_event_type = "new_tokens"
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.debug(err)
            cookie_headers_event_type = "refresh_failed"

        response = {
            "status": "307",
            "statusDescription": "Temporary Redirect",
            "headers": {
                "location": [{"key": "location", "value": redirected_from_uri}],
                "set-cookie": generate_cookie_headers(
                    cookie_headers_event_type,
                    CONFIG.get("client_id"),
                    CONFIG.get("oauth_scopes"),
                    tokens,
                    domain_name,
                    CONFIG.get("cookie_settings"),
                ),
                **CONFIG.get("cloud_front_headers", {}),
            },
        }
        # Redirect the user back to their requested uri
        # with new tokens at hand
        return response

    # Send a basic html error response and inform the user
    # why refresh was unsuccessful
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.info(err)
        LOGGER.info(traceback.print_exc())

        response = {
            "body": create_error_html(
                "Refresh issue",
                f"Your sign-in refresh failed due to a technical issue: {err}",
                redirected_from_uri,
                "Try Again",
            ),
            "status": "400",
            "headers": {
                "content-type": [
                    {"key": "Content-Type", "value": "text/html; charset=UTF-8"}
                ],
                **CONFIG.get("cloud_front_headers", {}),
            },
        }

        return response


def validate_refresh_request(current_nonce, original_nonce, tokens):
    """Validate that nonce and tokens are present.

    Args:
        current_nonce (str): The current nonce code.
        original_nonce (str): The original nonce code.
        tokens (Dict[str, str]): A dictionary of all the token_types
            and their corresponding token values (id, auth, refresh).

    """
    if not original_nonce:
        msg = (
            "Your browser didn't send the nonce cookie along, "
            "but it is required for security (prevent CSRF)."
        )
        LOGGER.error(msg)
        raise Exception(msg)

    if current_nonce != original_nonce:
        msg = "Nonce mismatch"
        LOGGER.error(msg)
        raise Exception(msg)

    for token_type, token in tokens.items():
        if not token:
            msg = f"Missing {token_type}"
            LOGGER.error(msg)
            raise Exception(msg)



"""CFNgin utilities."""
from __future__ import annotations

import copy
import locale
import logging
import os
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
import uuid
import zipfile
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    OrderedDict,
    Type,
    Union,
    cast,
)

import botocore.client
import botocore.exceptions
import dateutil
import yaml
from yaml.constructor import ConstructorError

from .awscli_yamlhelper import yaml_parse
from .session_cache import get_session

if TYPE_CHECKING:
    from mypy_boto3_route53.client import Route53Client
    from mypy_boto3_route53.type_defs import ResourceRecordSetTypeDef
    from mypy_boto3_s3.client import S3Client

    from ..config.models.cfngin import (
        CfnginPackageSourcesDefinitionModel,
        GitCfnginPackageSourceDefinitionModel,
        LocalCfnginPackageSourceDefinitionModel,
        S3CfnginPackageSourceDefinitionModel,
    )
    from .blueprints.base import Blueprint

LOGGER = logging.getLogger(__name__)


def camel_to_snake(name: str) -> str:
    """Convert CamelCase to snake_case.

    Args:
        name (str): The name to convert from CamelCase to snake_case.

    Returns:
        str: Converted string.

    """
    sub_str_1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", sub_str_1).lower()


def convert_class_name(kls: type) -> str:
    """Get a string that represents a given class.

    Args:
        kls: The class being analyzed for its name.

    Returns:
        The name of the given kls.

    """
    return camel_to_snake(kls.__name__)


def parse_zone_id(full_zone_id: str) -> str:
    """Parse the returned hosted zone id and returns only the ID itself."""
    return full_zone_id.split("/")[2]


def get_hosted_zone_by_name(client: Route53Client, zone_name: str) -> Optional[str]:
    """Get the zone id of an existing zone by name.

    Args:
        client: The connection used to interact with Route53's API.
        zone_name: The name of the DNS hosted zone to create.

    Returns:
        The Id of the Hosted Zone.

    """
    paginator = client.get_paginator("list_hosted_zones")

    for page in paginator.paginate():
        for zone in page["HostedZones"]:
            if zone["Name"] == zone_name:
                return parse_zone_id(zone["Id"])
    return None


def get_or_create_hosted_zone(client: Route53Client, zone_name: str) -> str:
    """Get the Id of an existing zone, or create it.

    Args:
        client: The connection used to interact with Route53's API.
        zone_name: The name of the DNS hosted zone to create.

    Returns:
        The Id of the Hosted Zone.

    """
    zone_id = get_hosted_zone_by_name(client, zone_name)
    if zone_id:
        return zone_id

    LOGGER.debug("zone %s does not exist; creating", zone_name)

    reference = uuid.uuid4().hex

    response = client.create_hosted_zone(Name=zone_name, CallerReference=reference)

    return parse_zone_id(response["HostedZone"]["Id"])


class SOARecordText:
    """Represents the actual body of an SOARecord."""

    def __init__(self, record_text: str) -> None:
        """Instantiate class."""
        (
            self.nameserver,
            self.contact,
            self.serial,
            self.refresh,
            self.retry,
            self.expire,
            self.min_ttl,
        ) = record_text.split()

    def __str__(self) -> str:
        """Convert an instance of this class to a string."""
        return " ".join(
            [
                self.nameserver,
                self.contact,
                self.serial,
                self.refresh,
                self.retry,
                self.expire,
                self.min_ttl,
            ]
        )


class SOARecord:
    """Represents an SOA record."""

    def __init__(self, record: ResourceRecordSetTypeDef) -> None:
        """Instantiate class."""
        self.name = record["Name"]
        self.text = SOARecordText(
            record.get("ResourceRecords", [{"Value": ""}])[0]["Value"]
        )
        self.ttl = record.get("TTL", 0)


def get_soa_record(client: Route53Client, zone_id: str, zone_name: str) -> SOARecord:
    """Get the SOA record for zone_name from zone_id.

    Args:
        client: The connection used to interact with Route53's API.
        zone_id: The AWS Route53 zone id of the hosted zone to query.
        zone_name: The name of the DNS hosted zone to create.

    Returns:
        An object representing the parsed SOA record returned from AWS Route53.

    """
    response = client.list_resource_record_sets(
        HostedZoneId=zone_id,
        StartRecordName=zone_name,
        StartRecordType="SOA",
        MaxItems="1",
    )
    return SOARecord(response["ResourceRecordSets"][0])


def create_route53_zone(client: Route53Client, zone_name: str) -> str:
    """Create the given zone_name if it doesn't already exists.

    Also sets the SOA negative caching TTL to something short (300 seconds).

    Args:
        client: The connection used to interact with Route53's API.
        zone_name: The name of the DNS hosted zone to create.

    Returns:
        The zone id returned from AWS for the existing, or newly created zone.

    """
    if not zone_name.endswith("."):
        zone_name += "."
    zone_id = get_or_create_hosted_zone(client, zone_name)
    old_soa = get_soa_record(client, zone_id, zone_name)

    # If the negative cache value is already 300, don't update it.
    if old_soa.text.min_ttl == "300":
        return zone_id

    new_soa = copy.deepcopy(old_soa)
    LOGGER.debug("updating negative caching value on zone %s to 300", zone_name)
    new_soa.text.min_ttl = "300"
    client.change_resource_record_sets(
        HostedZoneId=zone_id,
        ChangeBatch={
            "Comment": "Update SOA min_ttl to 300.",
            "Changes": [
                {
                    "Action": "UPSERT",
                    "ResourceRecordSet": {
                        "Name": zone_name,
                        "Type": "SOA",
                        "TTL": old_soa.ttl,
                        "ResourceRecords": [{"Value": str(new_soa.text)}],
                    },
                },
            ],
        },
    )
    return zone_id


def yaml_to_ordered_dict(
    stream: str,
    loader: Union[Type[yaml.Loader], Type[yaml.SafeLoader]] = yaml.SafeLoader,
) -> OrderedDict[str, Any]:
    """yaml.load alternative with preserved dictionary order.

    Args:
        stream: YAML string to load.
        loader: PyYAML loader class. Defaults to safe load.

    """

    class OrderedUniqueLoader(loader):  # type: ignore
        """Subclasses the given pyYAML `loader` class.

        Validates all sibling keys to insure no duplicates.

        Returns:
            OrderedDict: instead of a Dict.

        """

        # keys which require no duplicate siblings.
        NO_DUPE_SIBLINGS = ["stacks", "class_path"]
        # keys which require no duplicate children keys.
        NO_DUPE_CHILDREN = ["stacks"]

        @staticmethod
        def _error_mapping_on_dupe(
            node: Union[yaml.MappingNode, yaml.ScalarNode, yaml.SequenceNode],
            node_name: str,
        ) -> None:
            """Check mapping node for dupe children keys."""
            if isinstance(node, yaml.MappingNode):
                mapping: Dict[str, Any] = {}
                for val in node.value:
                    a = val[0]
                    b = mapping.get(a.value, None)
                    if b:
                        raise ConstructorError(
                            f"{node_name} mapping cannot have duplicate keys "
                            f"{b.start_mark} {a.start_mark}"
                        )
                    mapping[a.value] = a

        def _validate_mapping(
            self,
            node: Union[yaml.MappingNode, yaml.ScalarNode, yaml.SequenceNode],
            deep: bool = False,
        ) -> OrderedDict[Any, Any]:
            if not isinstance(node, yaml.MappingNode):
                raise ConstructorError(
                    None,
                    None,
                    f"expected a mapping node, but found {node.id}",
                    node.start_mark,
                )
            mapping: OrderedDict[Any, Any] = OrderedDict()
            for key_node, value_node in node.value:
                key = cast(object, self.construct_object(key_node, deep=deep))
                try:
                    hash(key)
                except TypeError as exc:
                    raise ConstructorError(
                        "while constructing a mapping",
                        node.start_mark,
                        f"found unhashable key ({exc})",
                        key_node.start_mark,
                    ) from exc
                # prevent duplicate sibling keys for certain "keywords".
                if key in mapping and key in self.NO_DUPE_SIBLINGS:
                    raise ConstructorError(
                        f"{key} key cannot have duplicate siblings "
                        f"{node.start_mark} {key_node.start_mark}"
                    )
                if key in self.NO_DUPE_CHILDREN:
                    # prevent duplicate children keys for this mapping.
                    self._error_mapping_on_dupe(value_node, key_node.value)
                value = cast(object, self.construct_object(value_node, deep=deep))
                mapping[key] = value
            return mapping

        def construct_mapping(
            self,
            node: Union[yaml.MappingNode, yaml.ScalarNode, yaml.SequenceNode],
            deep: bool = False,
        ) -> OrderedDict[Any, Any]:
            """Override parent method to use OrderedDict."""
            if isinstance(node, yaml.MappingNode):
                self.flatten_mapping(node)
            return self._validate_mapping(node, deep=deep)

        def construct_yaml_map(
            self, node: Union[yaml.MappingNode, yaml.ScalarNode, yaml.SequenceNode]
        ) -> Iterator[OrderedDict[Any, Any]]:
            data: OrderedDict[Any, Any] = OrderedDict()
            yield data
            value: OrderedDict[Any, Any] = self.construct_mapping(node)
            data.update(value)

    OrderedUniqueLoader.add_constructor(
        "tag:yaml.org,2002:map", OrderedUniqueLoader.construct_yaml_map
    )
    return yaml.load(stream, OrderedUniqueLoader)


def uppercase_first_letter(string_: str) -> str:
    """Return string with first character upper case."""
    return string_[0].upper() + string_[1:]


def cf_safe_name(name: str) -> str:
    """Convert a name to a safe string for a CloudFormation resource.

    Given a string, returns a name that is safe for use as a CloudFormation
    Resource. (ie: Only alphanumeric characters)

    """
    alphanumeric = r"[a-zA-Z0-9]+"
    parts = re.findall(alphanumeric, name)
    return "".join(uppercase_first_letter(part) for part in parts)


def read_value_from_path(value: str, *, root_path: Optional[Path] = None) -> str:
    """Enable translators to read values from files.

    The value can be referred to with the `file://` prefix.

    Example:
        ::

            conf_key: ${kms file://kms_value.txt}

    """
    if value.startswith("file://"):
        path = value.split("file://", 1)[1]
        if os.path.isabs(path):
            read_path = Path(path)
        else:
            root_path = root_path or Path.cwd()
            if root_path.is_dir():
                read_path = root_path / path
            else:
                read_path = root_path.parent / path
        if read_path.is_file():
            return read_path.read_text(
                encoding=locale.getpreferredencoding(do_setlocale=False)
            )
        if read_path.is_dir():
            raise ValueError(
                f"path must lead to a file not directory: {read_path.absolute()}"
            )
        raise ValueError(f"path does not exist: {read_path.absolute()}")
    return value


def get_client_region(client: Any) -> str:
    """Get the region from a boto3 client.

    Args:
        client: The client to get the region from.

    Returns:
        AWS region string.

    """
    return client._client_config.region_name  # type: ignore


def get_s3_endpoint(client: Any) -> str:
    """Get the s3 endpoint for the given boto3 client.

    Args:
        client: The client to get the endpoint from.

    Returns:
        The AWS endpoint for the client.

    """
    return client._endpoint.host  # type: ignore


def s3_bucket_location_constraint(region: Optional[str]) -> Optional[str]:
    """Return the appropriate LocationConstraint info for a new S3 bucket.

    When creating a bucket in a region OTHER than us-east-1, you need to
    specify a LocationConstraint inside the CreateBucketConfiguration argument.
    This function helps you determine the right value given a given client.

    Args:
        region: The region where the bucket will be created in.

    Returns:
        The string to use with the given client for creating a bucket.

    """
    if region == "us-east-1":
        return ""
    return region


def ensure_s3_bucket(
    s3_client: S3Client,
    bucket_name: str,
    bucket_region: Optional[str] = None,
    *,
    create: bool = True,
    persist_graph: bool = False,
) -> None:
    """Ensure an s3 bucket exists, if it does not then create it.

    Args:
        s3_client: An s3 client used to verify and create the bucket.
        bucket_name: The bucket being checked/created.
        bucket_region: The region to create the bucket in.
            If not provided, will be determined by s3_client's region.
        create: Whether to create the bucket if it does not exist.
        persist_graph: Check bucket for recommended settings.
            If creating a new bucket, it will be created with recommended
            settings.

    """
    try:
        s3_client.head_bucket(Bucket=bucket_name)
        if persist_graph:
            response = s3_client.get_bucket_versioning(Bucket=bucket_name)
            state = response.get("Status", "disabled").lower()
            if state != "enabled":
                LOGGER.warning(
                    'versioning is %s on bucket "%s"; it is recommended to '
                    "enable versioning when using persistent graphs",
                    state,
                    bucket_name,
                )
            if response.get("MFADelete", "Disabled") != "Disabled":
                LOGGER.warning(
                    'MFADelete must be disabled on bucket "%s" when using '
                    "persistent graphs to allow for proper management of "
                    "the graphs",
                    bucket_name,
                )
    except botocore.exceptions.ClientError as err:
        if err.response["Error"]["Message"] == "Not Found" and create:
            # can't use s3_client.exceptions.NoSuchBucket here.
            # it does not work if the bucket was recently deleted.
            LOGGER.debug("creating bucket %s", bucket_name)
            create_args: Dict[str, Any] = {"Bucket": bucket_name}
            location_constraint = s3_bucket_location_constraint(bucket_region)
            if location_constraint:
                create_args["CreateBucketConfiguration"] = {
                    "LocationConstraint": location_constraint
                }
            s3_client.create_bucket(**create_args)
            if persist_graph:
                s3_client.put_bucket_versioning(
                    Bucket=bucket_name, VersioningConfiguration={"Status": "Enabled"}
                )
            return
        if err.response["Error"]["Message"] == "Forbidden":
            LOGGER.exception(
                "Access denied for bucket %s. Did you remember "
                "to use a globally unique name?",
                bucket_name,
            )
        elif err.response["Error"]["Message"] != "Not Found":
            LOGGER.exception('error creating bucket "%s"', bucket_name)
        raise


def parse_cloudformation_template(template: str) -> Dict[str, Any]:
    """Parse CFN template string.

    Leverages the vendored aws-cli yamlhelper to handle JSON or YAML templates.

    Args:
        template: The template body.

    """
    return yaml_parse(template)


def is_within_directory(directory: Path | str, target: str) -> bool:
    """Check if file is in directory.

    Determines if the provided path is within a specific directory or its subdirectories.

    Args:
        directory (Union[Path, str]): Path of the directory we're checking.
        target (str): Path of the file we're checking for containment.

    Returns:
        bool: True if the target is in the directory or subdirectories, False otherwise.

    """
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)
    prefix = os.path.commonprefix([abs_directory, abs_target])
    return prefix == abs_directory


def safe_tar_extract(
    tar: tarfile.TarFile,
    path: Path | str = ".",
    members: list[tarfile.TarInfo] | None = None,
    *,
    numeric_owner: bool = False,
):
    """Safely extract the contents of a tar file to a specified directory.

    This code is modified from a PR provided to Runway project
    to address CVE-2007-4559.

    Args:
        tar (TarFile): The tar file object that will be extracted.
        path (Union[Path, str], optional): The directory to extract the tar into.
        members (List[TarInfo] | None, optional): List of TarInfo objects to extract.
        numeric_owner (bool, optional): Enable usage of owner and group IDs when extracting.

    Raises:
        Exception: If any tar file tries to go outside the specified area.

    """
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")
    tar.extractall(path, members, numeric_owner=numeric_owner)


class Extractor:
    """Base class for extractors."""

    extension: ClassVar[str] = ""

    def __init__(self, archive: Optional[Path] = None) -> None:
        """Instantiate class.

        Args:
            archive (str): Archive path.

        """
        self.archive = archive

    def set_archive(self, dir_name: Path) -> None:
        """Update archive filename to match directory name & extension.

        Args:
            dir_name: Archive directory name

        """
        self.archive = dir_name.with_suffix(self.extension)


class TarExtractor(Extractor):
    """Extracts tar archives."""

    extension: ClassVar[str] = ".tar"

    def extract(self, destination: Path) -> None:
        """Extract the archive."""
        with tarfile.open(self.archive, "r:") as tar:
            safe_tar_extract(tar, path=destination)


class TarGzipExtractor(Extractor):
    """Extracts compressed tar archives."""

    extension: ClassVar[str] = ".tar.gz"

    def extract(self, destination: Path) -> None:
        """Extract the archive."""
        with tarfile.open(self.archive, "r:gz") as tar:
            safe_tar_extract(tar, path=destination)


class ZipExtractor(Extractor):
    """Extracts zip archives."""

    extension: ClassVar[str] = ".zip"

    def extract(self, destination: Path) -> None:
        """Extract the archive."""
        if self.archive:
            with zipfile.ZipFile(self.archive, "r") as zip_ref:
                zip_ref.extractall(destination)


class SourceProcessor:
    """Makes remote python package sources available in current environment."""

    ISO8601_FORMAT = "%Y%m%dT%H%M%SZ"

    def __init__(
        self,
        sources: CfnginPackageSourcesDefinitionModel,
        cache_dir: Path,
    ) -> None:
        """Process a config's defined package sources.

        Args:
            sources: Package sources from CFNgin config dictionary.
            cache_dir: Path where remote sources will be cached.

        """
        self.cache_dir = cache_dir
        self.package_cache_dir = cache_dir / "packages"
        self.sources = sources
        self.configs_to_merge: List[Path] = []
        self.create_cache_directories()

    def create_cache_directories(self) -> None:
        """Ensure that SourceProcessor cache directories exist."""
        self.package_cache_dir.mkdir(parents=True, exist_ok=True)

    def get_package_sources(self) -> None:
        """Make remote python packages available for local use."""
        # Checkout local modules
        for config in self.sources.local:
            self.fetch_local_package(config=config)
        # Checkout S3 repositories specified in config
        for config in self.sources.s3:
            self.fetch_s3_package(config=config)
        # Checkout git repositories specified in config
        for config in self.sources.git:
            self.fetch_git_package(config=config)

    def fetch_local_package(
        self, config: LocalCfnginPackageSourceDefinitionModel
    ) -> None:
        """Make a local path available to current CFNgin config.

        Args:
            config: Package source config.

        """
        # Update sys.path & merge in remote configs (if necessary)
        self.update_paths_and_config(
            config=config, pkg_dir_name=config.source, pkg_cache_dir=Path.cwd()
        )

    def fetch_s3_package(self, config: S3CfnginPackageSourceDefinitionModel) -> None:
        """Make a remote S3 archive available for local use.

        Args:
            config: Package source config.

        """
        extractor_map = {
            ".tar.gz": TarGzipExtractor,
            ".tar": TarExtractor,
            ".zip": ZipExtractor,
        }
        extractor = None
        dir_name = ""
        for suffix, class_ in extractor_map.items():
            if config.key.endswith(suffix):
                extractor = class_()
                LOGGER.debug(
                    'using extractor %s for S3 object "%s" in bucket %s',
                    class_.__name__,
                    config.key,
                    config.bucket,
                )
                dir_name = self.sanitize_uri_path(
                    f"s3-{config.bucket}-{config.key[: -len(suffix)]}"
                )
                break

        if extractor is None:
            raise ValueError(
                f'Archive type could not be determined for S3 object "{config.key}" '
                f"in bucket {config.bucket}."
            )

        session = get_session(region=None)
        extra_s3_args: Dict[str, Any] = {}
        if config.requester_pays:
            extra_s3_args["RequestPayer"] = "requester"

        # We can skip downloading the archive if it's already been cached
        if config.use_latest:
            try:
                # LastModified should always be returned in UTC, but it doesn't
                # hurt to explicitly convert it to UTC again just in case
                modified_date = (
                    session.client("s3")
                    .head_object(Bucket=config.bucket, Key=config.key, **extra_s3_args)[
                        "LastModified"
                    ]  # type: ignore
                    .astimezone(dateutil.tz.tzutc())  # type: ignore
                )
            except botocore.exceptions.ClientError as client_error:
                LOGGER.error(
                    "error checking modified date of s3://%s/%s : %s",
                    config.bucket,
                    config.key,
                    client_error,
                )
                sys.exit(1)
            dir_name += f"-{modified_date.strftime(self.ISO8601_FORMAT)}"
        cached_dir_path = self.package_cache_dir / dir_name
        if not cached_dir_path.is_dir():
            LOGGER.debug(
                "remote package s3://%s/%s does not appear to have "
                "been previously downloaded; starting download and "
                "extraction to %s",
                config.bucket,
                config.key,
                cached_dir_path,
            )
            tmp_dir = tempfile.mkdtemp(prefix="cfngin")
            tmp_package_path = os.path.join(tmp_dir, dir_name)
            with tempfile.TemporaryDirectory(prefix="runway-cfngin") as tmp_dir:
                tmp_package_path = Path(tmp_dir) / dir_name
                extractor.set_archive(tmp_package_path)
                LOGGER.debug(
                    "starting remote package download from S3 to %s "
                    'with extra S3 options "%s"',
                    extractor.archive,
                    str(extra_s3_args),
                )
                session.resource("s3").Bucket(config.bucket).download_file(
                    Key=config.key,
                    Filename=str(extractor.archive),
                    ExtraArgs=extra_s3_args,
                )
                LOGGER.debug(
                    "download complete; extracting downloaded package to %s",
                    tmp_package_path,
                )
                extractor.extract(tmp_package_path)
                LOGGER.debug(
                    "moving extracted package directory %s to the "
                    "CFNgin cache at %s",
                    dir_name,
                    self.package_cache_dir,
                )
                shutil.move(str(tmp_package_path), self.package_cache_dir)
        else:
            LOGGER.debug(
                "remote package s3://%s/%s appears to have "
                "been previously downloaded to %s; download skipped",
                config.bucket,
                config.key,
                cached_dir_path,
            )

        # Update sys.path & merge in remote configs (if necessary)
        self.update_paths_and_config(config=config, pkg_dir_name=dir_name)

    def fetch_git_package(self, config: GitCfnginPackageSourceDefinitionModel) -> None:
        """Make a remote git repository available for local use.

        Args:
            config: Package source config.

        """
        # only loading git here when needed to avoid load errors on systems
        # without git installed
        from git.repo import Repo  # pylint: disable=import-outside-toplevel

        ref = self.determine_git_ref(config)
        dir_name = self.sanitize_git_path(uri=config.uri, ref=ref)
        cached_dir_path = self.package_cache_dir / dir_name

        # We can skip cloning the repo if it's already been cached
        if not cached_dir_path.is_dir():
            LOGGER.debug(
                "remote repo %s does not appear to have been "
                "previously downloaded; starting clone to %s",
                config.uri,
                cached_dir_path,
            )
            tmp_dir = tempfile.mkdtemp(prefix="cfngin")
            try:
                tmp_repo_path = os.path.join(tmp_dir, dir_name)
                with Repo.clone_from(config.uri, tmp_repo_path) as repo:
                    repo.head.set_reference(ref)
                    repo.head.reset(index=True, working_tree=True)
                shutil.move(tmp_repo_path, self.package_cache_dir)
            finally:
                shutil.rmtree(tmp_dir)
        else:
            LOGGER.debug(
                "remote repo %s appears to have been previously "
                "cloned to %s; download skipped",
                config.uri,
                cached_dir_path,
            )

        # Update sys.path & merge in remote configs (if necessary)
        self.update_paths_and_config(config=config, pkg_dir_name=dir_name)

    def update_paths_and_config(
        self,
        config: Union[
            GitCfnginPackageSourceDefinitionModel,
            LocalCfnginPackageSourceDefinitionModel,
            S3CfnginPackageSourceDefinitionModel,
        ],
        pkg_dir_name: str,
        pkg_cache_dir: Optional[Path] = None,
    ) -> None:
        """Handle remote source defined sys.paths & configs.

        Args:
            config: Package source config.
            pkg_dir_name: Directory name of the CFNgin archive.
            pkg_cache_dir: Fully qualified path to CFNgin cache cache directory.

        """
        if not pkg_cache_dir:
            pkg_cache_dir = self.package_cache_dir
        cached_dir_path = pkg_cache_dir / pkg_dir_name

        # Add the appropriate directory (or directories) to sys.path
        if config.paths:
            for path in config.paths:
                path_to_append = (cached_dir_path / path).resolve()
                LOGGER.debug("appending to python sys.path: %s", path_to_append)
                sys.path.append(str(path_to_append))
        else:
            sys.path.append(str(cached_dir_path.resolve()))

        # If the configuration defines a set of remote config yaml files to
        # include, add them to the list for merging
        if config.configs:
            for config_filename in config.configs:
                self.configs_to_merge.append(cached_dir_path / config_filename)

    @staticmethod
    def git_ls_remote(uri: str, ref: str) -> str:
        """Determine the latest commit id for a given ref.

        Args:
            uri: Git URI.
            ref: Git ref.

        """
        LOGGER.debug("getting commit ID from repo: %s", " ".join(uri))
        ls_remote_output = subprocess.check_output(["git", "ls-remote", uri, ref])
        if b"\t" in ls_remote_output:
            commit_id = ls_remote_output.split(b"\t", maxsplit=1)[0]
            LOGGER.debug("matching commit id found: %s", commit_id)
            return commit_id.decode()
        raise ValueError(f'Ref "{ref}" not found for repo {uri}.')

    @staticmethod
    def determine_git_ls_remote_ref(
        config: GitCfnginPackageSourceDefinitionModel,
    ) -> str:
        """Determine the ref to be used with the "git ls-remote" command.

        Args:
            config: Git package source config.

        Returns:
            A branch reference or "HEAD".

        """
        return f"refs/heads/{config.branch}" if config.branch else "HEAD"

    def determine_git_ref(self, config: GitCfnginPackageSourceDefinitionModel) -> str:
        """Determine the ref to be used for ``git checkout``.

        Args:
            config: Git package source config.

        Returns:
            A commit id or tag name.

        """
        if config.commit:
            return config.commit
        if config.tag:
            return config.tag
        return self.git_ls_remote(  # get a commit id to use
            config.uri, self.determine_git_ls_remote_ref(config)
        )

    @staticmethod
    def sanitize_uri_path(uri: str) -> str:
        """Take a URI and converts it to a directory safe path.

        Args:
            uri: URI to sanitize.

        Returns:
            Directory name for the supplied uri.

        """
        for i in ["@", "/", ":"]:
            uri = uri.replace(i, "_")
        return uri

    def sanitize_git_path(self, uri: str, ref: Optional[str] = None) -> str:
        """Take a git URI and ref and converts it to a directory safe path.

        Args:
            uri: Git URI. (e.g. ``git@github.com:foo/bar.git``)
            ref: Git ref to be appended to the path.

        Returns:
            Directory name for the supplied uri

        """
        if uri.endswith(".git"):
            dir_name = uri[:-4]  # drop .git
        else:
            dir_name = uri
        dir_name = self.sanitize_uri_path(dir_name)
        if ref is not None:
            dir_name += f"-{ref}"
        return dir_name


def stack_template_key_name(blueprint: Blueprint) -> str:
    """Given a blueprint, produce an appropriate key name.

    Args:
        blueprint: The blueprint object to create the key from.

    Returns:
        Key name resulting from blueprint.

    """
    name = blueprint.name
    return f"stack_templates/{blueprint.context.get_fqn(name)}/{name}-{blueprint.version}.json"



"""CFNgin graph action."""
from __future__ import annotations

import json
import logging
import sys
from typing import TYPE_CHECKING, Any, Iterable, List, TextIO, Tuple, Union

from ..plan import merge_graphs
from .base import BaseAction

if TYPE_CHECKING:
    from ..plan import Graph, Step

LOGGER = logging.getLogger(__name__)


def each_step(graph: Graph) -> Iterable[Tuple[Step, List[Step]]]:
    """Yield each step and it's direct dependencies.

    Args:
        graph: Graph to iterate over.

    """
    steps = graph.topological_sort()
    steps.reverse()

    for step in steps:
        deps = graph.downstream(step.name)
        yield step, deps


def dot_format(out: TextIO, graph: Graph, name: str = "digraph") -> None:
    """Output a graph using the graphviz "dot" format.

    Args:
        out: Where output will be written.
        graph: Graph to be output.
        name: Name of the graph.

    """
    out.write(f"digraph {name} {{\n")
    for step, deps in each_step(graph):
        for dep in deps:
            out.write(f'  "{step}" -> "{dep}";\n')

    out.write("}\n")


def json_format(out: TextIO, graph: Graph) -> None:
    """Output the graph in a machine readable JSON format.

    Args:
        out: Where output will be written.
        graph: Graph to be output.

    """
    steps = {
        step.name: {"deps": [dep.name for dep in deps]}
        for step, deps in each_step(graph)
    }

    json.dump({"steps": steps}, out, indent=4)
    out.write("\n")


FORMATTERS = {
    "dot": dot_format,
    "json": json_format,
}


class Action(BaseAction):
    """Responsible for outputting a graph for the current CFNgin config."""

    DESCRIPTION = "Print graph"
    NAME = "graph"

    @property
    def _stack_action(self) -> Any:
        """Run against a step."""
        return None

    def run(
        self,
        *,
        concurrency: int = 0,  # pylint: disable=unused-argument
        dump: Union[bool, str] = False,  # pylint: disable=unused-argument
        force: bool = False,  # pylint: disable=unused-argument
        outline: bool = False,  # pylint: disable=unused-argument
        tail: bool = False,  # pylint: disable=unused-argument
        upload_disabled: bool = False,  # pylint: disable=unused-argument
        **kwargs: Any,
    ) -> None:
        """Generate the underlying graph and prints it."""
        graph = self._generate_plan(
            require_unlocked=False, include_persistent_graph=True
        ).graph
        if self.context.persistent_graph:
            graph = merge_graphs(self.context.persistent_graph, graph)
        if kwargs.get("reduce"):
            # This will perform a a transitive reduction on the underlying
            # graph, producing less edges. Mostly useful for the "dot" format,
            # when converting to PNG, so it creates a prettier/cleaner
            # dependency graph.
            graph.transitive_reduction()

        fn = FORMATTERS[str(kwargs.get("format", "json"))]
        fn(sys.stdout, graph)
        sys.stdout.flush()



"""CFNgin Blueprint type definitions."""
from __future__ import annotations

from typing import Any, Callable, List

from typing_extensions import TypedDict


class _RequiredBlueprintVariableTypeDef(TypedDict, total=False):
    """Type definition for runway.cfngin.blueprints.base.Blueprint.VARIABLES items."""

    type: Any


class _OptionalBlueprintVariableTypeDef(TypedDict, total=False):
    """Type definition for runway.cfngin.blueprints.base.Blueprint.VARIABLES items."""

    allowed_pattern: str
    allowed_values: List[Any]
    constraint_description: str
    default: Any
    description: str
    max_length: int
    max_value: int
    min_length: int
    min_value: int
    no_echo: bool
    validator: Callable[[Any], Any]


class BlueprintVariableTypeDef(
    _RequiredBlueprintVariableTypeDef, _OptionalBlueprintVariableTypeDef
):
    """Type definition for :attr:`runway.cfngin.blueprints.base.Blueprint.VARIABLES` items.

    Attributes:
        allowed_pattern: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            A regular expression that represents the patterns you want to allow
            for the Cloudformation Parameter.
        allowed_values: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            The values that should be allowed for the CloudFormation Parameter.
        constraint_description: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            A string that explains the constraint when the constraint
            is violated for the CloudFormation Parameter.
        default: The default value that should be used for the variable if none is
                provided in the config.
        description: A string that describes the purpose of the variable.
        max_length: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            The maximum length of the value for the CloudFormation Parameter.
        max_value: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            The max value for the CloudFormation Parameter.
        min_length: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            The minimum length of the value for the CloudFormation Parameter.
        min_value: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            The minimum value for the CloudFormation Parameter.
        no_echo: Only valid for variables whose type subclasses
            :class:`~runway.cfngin.blueprints.variables.types.CFNType`.
            Whether to mask the parameter value whenever anyone makes a call that
            describes the stack. If you set the value to true, the parameter
            value is masked with asterisks (*****).
        type: The type for the variable value. This can either be a native python
            type or one of the CFNgin variable types.
        validator: An optional function that can do custom validation of the variable.
            A validator function should take a single argument, the value being
            validated, and should return the value if validation is successful.
            If there is an issue validating the value, an exception
            (``ValueError``, ``TypeError``, etc) should be raised by the function.

    """  # noqa



"""Class mixins."""
from __future__ import annotations

import logging
import platform
import shutil
import subprocess
from contextlib import suppress
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
    cast,
    overload,
)

from typing_extensions import Literal

from .compat import shlex_join

if TYPE_CHECKING:
    from pathlib import Path

    from ._logging import RunwayLogger
    from .context import CfnginContext, RunwayContext

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class CliInterfaceMixin:
    """Mixin for adding CLI interface methods."""

    EXECUTABLE: ClassVar[str]
    """CLI executable."""

    ctx: Union[CfnginContext, RunwayContext]
    """CFNgin or Runway context object."""

    cwd: Path
    """Working directory where commands will be run."""

    @staticmethod
    def convert_to_cli_arg(arg_name: str, *, prefix: str = "--") -> str:
        """Convert string kwarg name into a CLI argument."""
        return f"{prefix}{arg_name.replace('_', '-')}"

    @classmethod
    def found_in_path(cls) -> bool:
        """Determine if executable is found in $PATH."""
        if shutil.which(cls.EXECUTABLE):
            return True
        return False

    @classmethod
    def generate_command(
        cls,
        command: Union[List[str], str],
        **kwargs: Optional[Union[bool, Iterable[str], str]],
    ) -> List[str]:
        """Generate command to be executed and log it.

        Args:
            command: Command to run.
            args: Additional args to pass to the command.

        Returns:
            The full command to be passed into a subprocess.

        """
        cmd = [cls.EXECUTABLE, *(command if isinstance(command, list) else [command])]
        cmd.extend(cls._generate_command_handle_kwargs(**kwargs))
        LOGGER.debug("generated command: %s", cls.list2cmdline(cmd))
        return cmd

    @classmethod
    def _generate_command_handle_kwargs(
        cls, **kwargs: Optional[Union[bool, Iterable[str], str]]
    ) -> List[str]:
        """Handle kwargs passed to generate_command."""
        result: List[str] = []
        for k, v in kwargs.items():
            if isinstance(v, str):
                result.extend([cls.convert_to_cli_arg(k), v])
            elif isinstance(v, (list, set, tuple)):
                for i in cast(Iterable[str], v):
                    result.extend([cls.convert_to_cli_arg(k), i])
            elif isinstance(v, bool) and v:
                result.append(cls.convert_to_cli_arg(k))
        return result

    @staticmethod
    def list2cmdline(split_command: Iterable[str]) -> str:
        """Combine a list of strings into a string that can be run as a command.

        Handles multi-platform differences.

        """
        if platform.system() == "Windows":
            return subprocess.list2cmdline(split_command)
        return shlex_join(split_command)

    @overload
    def _run_command(
        self,
        command: Union[Iterable[str], str],
        *,
        env: Optional[Dict[str, str]] = ...,
        suppress_output: Literal[True] = ...,
    ) -> str:
        ...

    @overload
    def _run_command(
        self,
        command: Union[Iterable[str], str],
        *,
        env: Optional[Dict[str, str]] = ...,
        suppress_output: Literal[False] = ...,
    ) -> None:
        ...

    def _run_command(
        self,
        command: Union[Iterable[str], str],
        *,
        env: Optional[Dict[str, str]] = None,
        suppress_output: bool = True,
    ) -> Optional[str]:
        """Run command.

        Args:
            command: Command to pass to shell to execute.
            env: Environment variables.
            suppress_output: If ``True``, the output of the subprocess written
                to ``sys.stdout`` and ``sys.stderr`` will be captured and
                returned as a string instead of being being written directly.

        """
        cmd_str = command if isinstance(command, str) else self.list2cmdline(command)
        LOGGER.verbose("running command: %s", cmd_str)
        if suppress_output:
            return subprocess.check_output(
                cmd_str,
                cwd=self.cwd,
                env=env or self.ctx.env.vars,
                shell=True,
                stderr=subprocess.PIPE,
                text=True,
            )
        subprocess.check_call(
            cmd_str,
            cwd=self.cwd,
            env=env or self.ctx.env.vars,
            shell=True,
        )
        return None


class DelCachedPropMixin:
    """Mixin to handle safely clearing the value of :func:`functools.cached_property`."""

    def _del_cached_property(self, *names: str) -> None:
        """Delete the cached value of a :func:`functools.cached_property`.

        Args:
            names: Names of the attribute that is cached. Can provide one or multiple.

        """
        for name in names:
            with suppress(AttributeError):
                delattr(self, name)



"""Response data models."""
from typing import List, Optional

from pydantic import Extra

from runway.utils import BaseModel


class AwsLambdaHookDeployResponse(BaseModel):
    """Data model for AwsLambdaHook deploy response.

    When returned by the hook as ``hook_data``, this model is dumped to a
    standard :class:`~typing.Dict` using the field's aliases as the ``key`` in
    place of the attribute names.
    This is done so that the ``key`` is a direct match to a CloudFormation
    Property where the value should be used.

    """

    bucket_name: str
    """Name of the S3 Bucket where the deployment package is located. (alias ``S3Bucket``)"""

    code_sha256: str
    """SHA256 of the deployment package.
    This can be used by CloudFormation as the value of ``AWS::Lambda::Version.CodeSha256``.
    (alias ``CodeSha256``)

    """

    compatible_architectures: Optional[List[str]] = None
    """A list of compatible instruction set architectures.
    (https://docs.aws.amazon.com/lambda/latest/dg/foundation-arch.html)
    (alias ``CompatibleArchitectures``)

    """

    compatible_runtimes: Optional[List[str]] = None
    """A list of compatible function runtimes.
    Used for filtering with ``ListLayers`` and ``ListLayerVersions``.
    (alias ``CompatibleRuntimes``)

    """

    license: Optional[str] = None
    """The layer's software license (alias ``License``). Can be any of the following:

    - A SPDX license identifier (e.g. ``MIT``).
    - The URL of a license hosted on the internet (e.g.
      ``https://opensource.org/licenses/MIT``).
    - The full text of the license.

    """

    object_key: str
    """Key (file path) of the deployment package S3 Object. (alias ``S3Key``)"""

    object_version_id: Optional[str] = None
    """The version ID of the deployment package S3 Object.
    This will only have a value if the S3 Bucket has versioning enabled.
    (alias ``S3ObjectVersion``)

    """

    runtime: str
    """Runtime of the Lambda Function. (alias ``Runtime``)"""

    class Config:
        """Model configuration."""

        allow_population_by_field_name = True
        extra = Extra.forbid
        fields = {
            "bucket_name": {"alias": "S3Bucket"},
            "code_sha256": {"alias": "CodeSha256"},
            "compatible_architectures": {"alias": "CompatibleArchitectures"},
            "compatible_runtimes": {"alias": "CompatibleRuntimes"},
            "license": {"alias": "License"},
            "object_key": {"alias": "S3Key"},
            "object_version_id": {"alias": "S3ObjectVersion"},
            "runtime": {"alias": "Runtime"},
        }



"""Hooks for AWS Lambda."""
from . import (
    base_classes,
    constants,
    deployment_package,
    docker,
    exceptions,
    models,
    python_requirements,
    source_code,
    type_defs,
)
from ._python_hooks import PythonFunction, PythonLayer

__all__ = [
    "PythonFunction",
    "PythonLayer",
    "base_classes",
    "constants",
    "deployment_package",
    "docker",
    "exceptions",
    "models",
    "python_requirements",
    "source_code",
    "type_defs",
]



"""CFNgin session caching."""
import logging
from typing import Optional

import boto3

from ..aws_sso_botocore.session import Session
from ..constants import BOTO3_CREDENTIAL_CACHE
from .ui import ui

LOGGER = logging.getLogger(__name__)

DEFAULT_PROFILE = None


def get_session(
    region: Optional[str] = None,
    profile: Optional[str] = None,
    access_key: Optional[str] = None,
    secret_key: Optional[str] = None,
    session_token: Optional[str] = None,
) -> boto3.Session:
    """Create a thread-safe boto3 session.

    Args:
        region: The region for the session.
        profile: The profile for the session.
        access_key: AWS Access Key ID.
        secret_key: AWS secret Access Key.
        session_token: AWS session token.

    Returns:
        A thread-safe boto3 session.

    """
    if profile:
        LOGGER.debug(
            'building session using profile "%s" in region "%s"',
            profile,
            region or "default",
        )
    elif access_key:
        LOGGER.debug(
            'building session with Access Key "%s" in region "%s"',
            access_key,
            region or "default",
        )

    session = boto3.Session(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token,
        botocore_session=Session(),  # type: ignore
        region_name=region,
        profile_name=profile,
    )
    cred_provider = session._session.get_component("credential_provider")  # type: ignore
    provider = cred_provider.get_provider("assume-role")  # type: ignore
    provider.cache = BOTO3_CREDENTIAL_CACHE
    provider._prompter = ui.getpass
    return session



"""``runway tfenv`` command group."""
# docs: file://./../../../../docs/source/commands.rst
from typing import Any

import click

from ... import options
from ._install import install
from ._list import list_installed
from ._run import run
from ._uninstall import uninstall

__all__ = ["install", "list_installed", "run", "uninstall"]

COMMANDS = [install, list_installed, run, uninstall]


@click.group("tfenv", short_help="terraform (install|run)")
@options.debug
@options.no_color
@options.verbose
def tfenv(**_: Any) -> None:
    """Terraform version management and execution.

    Runway's built-in Terraform version management allows for long-term
    stability of Terraform executions. Define a ".terraform-version" file
    in your Terraform module and that version will be automatically
    downloaded & used during Runway operations.

    """


for cmd in COMMANDS:  # register commands
    tfenv.add_command(cmd)



"""pip CLI interface."""
from __future__ import annotations

import logging
import re
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Tuple, Union, cast

from typing_extensions import Final, Literal

from ..compat import cached_property, shlex_join
from ..exceptions import RunwayError
from ..utils import Version
from .base_classes import DependencyManager

if TYPE_CHECKING:
    from _typeshed import StrPath

    from .._logging import RunwayLogger


LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class PipInstallFailedError(RunwayError):
    """Pip install failed."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate class. All args/kwargs are passed to parent method."""
        self.message = (
            "pip failed to install dependencies; "
            "review pip's output above to troubleshoot"
        )
        super().__init__(*args, **kwargs)


class Pip(DependencyManager):
    """pip CLI interface."""

    CONFIG_FILES: Final[Tuple[Literal["requirements.txt"]]] = ("requirements.txt",)
    """Configuration files used by pip."""

    EXECUTABLE: Final[Literal["pip"]] = "pip"
    """CLI executable."""

    @cached_property
    def python_version(self) -> Version:
        """Python version where pip is installed (``<major>.<minor>`` only)."""
        cmd_output = self._run_command([self.EXECUTABLE, "--version"])
        match = re.search(r"^pip \S* from .+ \(python (?P<version>\S*)\)$", cmd_output)
        if not match:
            LOGGER.warning(
                "unable to parse Python version from output:\n%s", cmd_output
            )
            return Version("0.0.0")
        return Version(match.group("version"))

    @cached_property
    def version(self) -> Version:
        """pip version."""
        cmd_output = self._run_command([self.EXECUTABLE, "--version"])
        match = re.search(r"^pip (?P<version>\S*) from .+$", cmd_output)
        if not match:
            LOGGER.warning("unable to parse pip version from output:\n%s", cmd_output)
            return Version("0.0.0")
        return Version(match.group("version"))

    @classmethod
    def dir_is_project(cls, directory: StrPath, **kwargs: Any) -> bool:
        """Determine if the directory contains a project for this dependency manager.

        Args:
            directory: Directory to check.

        """
        kwargs.setdefault("file_name", cls.CONFIG_FILES[0])
        requirements_txt = Path(directory) / kwargs["file_name"]
        if requirements_txt.is_file():
            return True
        return False

    @classmethod
    def generate_install_command(
        cls,
        *,
        cache_dir: Optional[StrPath] = None,
        no_cache_dir: bool = False,
        no_deps: bool = False,
        requirements: StrPath,
        target: StrPath,
    ) -> List[str]:
        """Generate the command that when run will install dependencies.

        This method is exposed to easily format the command to be run by with
        a subprocess or within a Docker container.

        Args:
            cache_dir: Store the cache data in the provided directory.
            no_cache_dir: Disable the cache.
            no_deps: Don't install package dependencies.
            requirements: Path to a ``requirements.txt`` file.
            target: Path to a directory where dependencies will be installed.

        """
        return cls.generate_command(
            "install",
            cache_dir=str(cache_dir) if cache_dir else None,
            disable_pip_version_check=True,
            no_cache_dir=no_cache_dir,
            no_deps=no_deps,
            no_input=True,
            requirement=str(requirements),
            target=str(target),
        )

    def install(
        self,
        *,
        cache_dir: Optional[StrPath] = None,
        extend_args: Optional[List[str]] = None,
        no_cache_dir: bool = False,
        no_deps: bool = False,
        requirements: StrPath,
        target: StrPath,
    ) -> Path:
        """Install dependencies to a target directory.

        Args:
            cache_dir: Store the cache data in the provided directory.
            extend_args: Optional list of extra arguments to pass to ``pip install``.
                This value will not be parsed or sanitized in any way - it will
                be used as is. It is the user's responsibility to ensure that
                there are no overlapping arguments between this list and the
                arguments that are automatically generated.
            no_cache_dir: Disable the cache.
            no_deps: Don't install package dependencies.
            requirements: Path to a ``requirements.txt`` file.
            target: Path to a directory where dependencies will be installed.

        Raises:
            PipInstallFailedError: The subprocess used to run the commend
                exited with an error.

        """
        target = Path(target) if not isinstance(target, Path) else target
        target.mkdir(exist_ok=True, parents=True)
        try:
            self._run_command(
                self.generate_install_command(
                    cache_dir=cache_dir,
                    no_cache_dir=no_cache_dir,
                    no_deps=no_deps,
                    requirements=requirements,
                    target=target,
                )
                + (extend_args or []),
                suppress_output=False,
            )
        except subprocess.CalledProcessError as exc:
            raise PipInstallFailedError from exc
        return target

    @classmethod
    def generate_command(
        cls,
        command: Union[List[str], str],
        **kwargs: Optional[Union[bool, Iterable[str], str]],
    ) -> List[str]:
        """Generate command to be executed and log it.

        Args:
            command: Command to run.
            args: Additional args to pass to the command.

        Returns:
            The full command to be passed into a subprocess.

        """
        cmd = [
            "python",
            "-m",
            cls.EXECUTABLE,
            *(command if isinstance(command, list) else [command]),
        ]
        cmd.extend(cls._generate_command_handle_kwargs(**kwargs))
        LOGGER.debug("generated command: %s", shlex_join(cmd))
        return cmd



"""List the versions of Terraform that have been installed by Runway and/or tfenv."""
# docs: file://./../../../../docs/source/commands.rst
import logging
from typing import Any

import click

from ....env_mgr.tfenv import TFEnvManager
from ... import options

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("list", short_help="list installed versions")
@options.debug
@options.no_color
@options.verbose
def list_installed(**_: Any) -> None:
    """List the versions of Terraform that have been installed by Runway and/or tfenv."""
    tfenv = TFEnvManager()
    versions = list(tfenv.list_installed())
    versions.sort()

    if versions:
        LOGGER.info("Terraform versions installed:")
        click.echo("\n".join(v.name for v in versions))
    else:
        LOGGER.warning(
            "no versions of Terraform installed at path %s", tfenv.versions_dir
        )



"""Sample app."""
from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3
from aws_cdk import core


class HelloConstruct(core.Construct):
    """Hello construct."""

    @property
    def buckets(self):
        """Return a tuple of buckets."""
        return tuple(self._buckets)

    def __init__(self, scope: core.Construct, id: str, num_buckets: int) -> None:
        """Instantiate class."""
        super().__init__(scope, id)
        self._buckets = []
        for i in range(0, num_buckets):
            self._buckets.append(s3.Bucket(self, f"Bucket-{i}"))

    def grant_read(self, principal: iam.IPrincipal):
        """Grant reed access."""
        for b in self.buckets:
            b.grant_read(principal, "*")



"""Retrieve a value from AWS Elastic Container Registry (ECR)."""
from __future__ import annotations

import base64
import logging
from typing import TYPE_CHECKING, Any, Union

from typing_extensions import Final, Literal

from ...lookups.handlers.base import LookupHandler

if TYPE_CHECKING:
    from mypy_boto3_ecr.client import ECRClient

    from ...context import CfnginContext, RunwayContext

LOGGER = logging.getLogger(__name__)


class EcrLookup(LookupHandler):
    """ECR Lookup."""

    TYPE_NAME: Final[Literal["ecr"]] = "ecr"
    """Name that the Lookup is registered as."""

    @staticmethod
    def get_login_password(client: ECRClient) -> str:
        """Get a password to login to ECR registry."""
        auth = client.get_authorization_token().get("authorizationData", [None])[0]
        if not auth or "authorizationToken" not in auth:
            raise ValueError("get_authorization_token did not return authorizationData")
        auth_token = base64.b64decode(auth["authorizationToken"]).decode()
        _, password = auth_token.split(":")
        return password

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls,
        value: str,
        context: Union[CfnginContext, RunwayContext],
        *__args: Any,
        **__kwargs: Any,
    ) -> Any:
        """Retrieve a value from AWS Elastic Container Registry (ECR).

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        """
        query, args = cls.parse(value)

        session = context.get_session(region=args.get("region"))
        client = session.client("ecr")

        if query == "login-password":
            result = cls.get_login_password(client)
        else:
            raise ValueError(f"ecr lookup does not support '{query}'")
        return cls.format_results(result, **args)



"""Purge all images from an ECR repository."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List

from ....utils import BaseModel

if TYPE_CHECKING:
    from mypy_boto3_ecr.client import ECRClient
    from mypy_boto3_ecr.type_defs import ImageIdentifierTypeDef

    from ....context import CfnginContext

LOGGER = logging.getLogger(__name__.replace("._", "."))


class HookArgs(BaseModel):
    """Hook arguments for ``purge_repository``."""

    repository_name: str
    """Name of the repository to purge."""


def delete_ecr_images(
    client: ECRClient,
    image_ids: List[ImageIdentifierTypeDef],
    repository_name: str,
) -> None:
    """Delete images from an ECR repository."""
    response = client.batch_delete_image(
        repositoryName=repository_name, imageIds=image_ids
    )
    if "failures" in response and response["failures"]:
        for msg in response["failures"]:
            LOGGER.info(
                "failed to delete image %s: (%s) %s",
                msg.get("imageId", {}).get("imageDigest")
                or msg.get("imageId", {}).get("imageTag"),
                msg.get("failureCode"),
                msg.get("failureReason"),
            )
        raise ValueError("failures present in response")


def list_ecr_images(
    client: ECRClient, repository_name: str
) -> List[ImageIdentifierTypeDef]:
    """List all images in an ECR repository."""
    image_ids: List[ImageIdentifierTypeDef] = []
    try:
        response = client.list_images(
            repositoryName=repository_name, filter={"tagStatus": "ANY"}
        )
        image_ids.extend(response["imageIds"])
        while response.get("nextToken"):
            response = client.list_images(
                filter={"tagStatus": "ANY"},
                nextToken=response["nextToken"],
                repositoryName=repository_name,
            )
            image_ids.extend(response["imageIds"])
        return [
            {"imageDigest": digest}
            for digest in {
                image["imageDigest"] for image in image_ids if image.get("imageDigest")
            }
        ]
    except client.exceptions.RepositoryNotFoundException:
        LOGGER.info("repository %s does not exist", repository_name)
        return []


def purge_repository(
    context: CfnginContext, *__args: Any, **kwargs: Any
) -> Dict[str, str]:
    """Purge all images from an ECR repository.

    Args:
        context: CFNgin context object.

    """
    args = HookArgs.parse_obj(kwargs)
    client = context.get_session().client("ecr")
    image_ids = list_ecr_images(client, repository_name=args.repository_name)
    if not image_ids:
        LOGGER.info("no images found in repository %s", args.repository_name)
        return {"status": "skipped"}
    delete_ecr_images(client, image_ids=image_ids, repository_name=args.repository_name)
    LOGGER.info("purged all images from repository")
    return {"status": "success"}



"""Docker image build hook.

Replicates the functionality of the ``docker image build`` CLI command.

"""
# pylint: disable=no-self-argument
from __future__ import annotations

import logging
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from docker.models.images import Image
from pydantic import DirectoryPath, Field, validator

from .....utils import BaseModel
from ..data_models import (
    DockerImage,
    ElasticContainerRegistry,
    ElasticContainerRegistryRepository,
)
from ..hook_data import DockerHookData

if TYPE_CHECKING:
    from .....context import CfnginContext

LOGGER = logging.getLogger(__name__.replace("._", "."))


class DockerImageBuildApiOptions(BaseModel):
    """Options for controlling Docker."""

    buildargs: Dict[str, Any] = {}
    """Dict of build-time variables that will be passed to Docker."""

    custom_context: bool = False
    """Whether to use custom context when providing a file object."""

    extra_hosts: Dict[str, str] = {}
    """Extra hosts to add to `/etc/hosts` in the build containers.
    Defined as a mapping of hostname to IP address.

    """

    forcerm: bool = False
    """Always remove intermediate containers, even after unsuccessful builds."""

    isolation: Optional[str] = None
    """Isolation technology used during build."""

    network_mode: Optional[str] = None
    """Network mode for the run commands during build."""

    nocache: bool = False
    """Whether to use cache."""

    platform: Optional[str] = None
    """Set platform if server is multi-platform capable.
    Uses format ``os[/arch[/variant]]``.

    """

    pull: bool = False
    """Whether to download any updates to the FROM image in the Dockerfile."""

    rm: bool = True
    """Whether to remove intermediate containers."""

    squash: bool = False
    """Whether to squash the resulting image layers into a single layer."""

    tag: Optional[str] = None
    """Optional name and tag to apply to the base image when it is built."""

    target: Optional[str] = None
    """Name of the build-stage to build in a multi-stage Dockerfile."""

    timeout: Optional[int] = None
    """HTTP timeout."""

    use_config_proxy: bool = False
    """If ``True`` and if the Docker client configuration file
    (``~/.docker/config.json`` by default) contains a proxy configuration,
    the corresponding environment variables will be set in the container
    being built.

    """


class ImageBuildArgs(BaseModel):
    """Args passed to image.build."""

    _ctx: Optional[CfnginContext] = Field(default=None, alias="context", export=False)

    ecr_repo: Optional[ElasticContainerRegistryRepository] = None  # depends on _ctx
    """AWS Elastic Container Registry repository information.
    Providing this will automatically construct the repo URI.
    If provided, do not provide ``repo``.

    If using a private registry, only ``repo_name`` is required.
    If using a public registry, ``repo_name`` and ``registry_alias``.

    """

    path: DirectoryPath = Path.cwd()
    """Path to the directory containing the Dockerfile."""

    dockerfile: str = "Dockerfile"  # depends on path for validation
    """Path within the build context to the Dockerfile."""

    repo: Optional[str] = None  # depends on ecr_repo
    """URI of a non Docker Hub repository where the image will be stored."""

    docker: DockerImageBuildApiOptions = DockerImageBuildApiOptions()  # depends on repo
    """Options for ``docker image build``."""

    tags: List[str] = ["latest"]
    """List of tags to apply to the image."""

    @validator("docker", pre=True, always=True, allow_reuse=True)
    def _set_docker(
        cls,
        v: Union[Dict[str, Any], DockerImageBuildApiOptions, Any],
        values: Dict[str, Any],
    ) -> Any:
        """Set the value of ``docker``."""
        repo = values["repo"]
        if repo:
            if isinstance(v, dict):
                v.setdefault("tag", repo)
            elif isinstance(v, DockerImageBuildApiOptions) and not v.tag:
                v.tag = repo
        return v

    @validator("ecr_repo", pre=True, allow_reuse=True)
    def _set_ecr_repo(cls, v: Any, values: Dict[str, Any]) -> Any:
        """Set the value of ``ecr_repo``."""
        if v and isinstance(v, dict):
            return ElasticContainerRegistryRepository.parse_obj(
                {
                    "repo_name": v.get("repo_name"),
                    "registry": ElasticContainerRegistry.parse_obj(
                        {
                            "account_id": v.get("account_id"),
                            "alias": v.get("registry_alias"),
                            "aws_region": v.get("aws_region"),
                            "context": values.get("context"),
                        }
                    ),
                }
            )
        return v

    @validator("repo", pre=True, always=True, allow_reuse=True)
    def _set_repo(cls, v: Optional[str], values: Dict[str, Any]) -> Optional[str]:
        """Set the value of ``repo``."""
        if v:
            return v

        ecr_repo: Optional[ElasticContainerRegistryRepository] = values.get("ecr_repo")
        if ecr_repo:
            return ecr_repo.fqn

        return None

    @validator("dockerfile", pre=True, always=True, allow_reuse=True)
    def _validate_dockerfile(cls, v: Any, values: Dict[str, Any]) -> Any:
        """Validate ``dockerfile``."""
        path: Path = values["path"]
        dockerfile = path / v
        if not dockerfile.is_file():
            raise ValueError(
                f"Dockerfile does not exist at path provided: {dockerfile}"
            )
        return v


def build(*, context: CfnginContext, **kwargs: Any) -> DockerHookData:
    """Docker image build hook.

    Replicates the functionality of ``docker image build`` CLI command.

    kwargs are parsed by :class:`~runway.cfngin.hooks.docker.image.ImageBuildArgs`.

    """
    args = ImageBuildArgs.parse_obj({"context": context, **kwargs})
    docker_hook_data = DockerHookData.from_cfngin_context(context)
    image, logs = cast(
        Tuple[Image, Iterator[Dict[str, str]]],
        docker_hook_data.client.images.build(path=str(args.path), **args.docker.dict()),
    )
    for msg in logs:  # iterate through JSON log messages
        if "stream" in msg:  # log if they contain a message
            LOGGER.info(msg["stream"].strip())  # remove any new line characters
    for tag in args.tags:
        image.tag(args.repo, tag=tag)
    image.reload()
    LOGGER.info("created image %s with tags %s", image.short_id, ", ".join(image.tags))
    docker_hook_data.image = DockerImage(image=image)
    return docker_hook_data.update_context(context)



"""Deployment package."""
from __future__ import annotations

import base64
import hashlib
import logging
import mimetypes
import stat
import sys
import zipfile
from contextlib import suppress
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)
from urllib.parse import urlencode

from typing_extensions import Final, Literal

from ....compat import cached_property
from ....core.providers.aws.s3 import Bucket
from ....core.providers.aws.s3.exceptions import (
    BucketAccessDeniedError,
    BucketNotFoundError,
    S3ObjectDoesNotExistError,
)
from ....exceptions import RequiredTagNotFoundError
from ....mixins import DelCachedPropMixin
from ....utils import FileHash
from .base_classes import Project
from .exceptions import DeploymentPackageEmptyError
from .models.args import AwsLambdaHookArgs

if TYPE_CHECKING:
    from pathlib import Path

    import igittigitt
    from mypy_boto3_s3.type_defs import HeadObjectOutputTypeDef, PutObjectOutputTypeDef

    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


_ProjectTypeVar = TypeVar("_ProjectTypeVar", bound=Project[AwsLambdaHookArgs])


class DeploymentPackage(DelCachedPropMixin, Generic[_ProjectTypeVar]):
    """AWS Lambda Deployment Package.

    When interacting with subclass of this instance, it is recommended to
    only call the methods defined within this parent class. This ensures
    compatibility with the S3 object class that can be returned.

    """

    META_TAGS: ClassVar[Dict[str, str]] = {
        "code_sha256": "runway.cfngin:awslambda.code_sha256",
        "compatible_architectures": "runway.cfngin:awslambda.compatible_architectures",
        "compatible_runtimes": "runway.cfngin:awslambda.compatible_runtimes",
        "license": "runway.cfngin:awslambda.license",
        "md5_checksum": "runway.cfngin:awslambda.md5_checksum",
        "runtime": "runway.cfngin:awslambda.runtime",
        "source_code.hash": "runway.cfngin:awslambda.source_code.hash",
    }
    """Mapping of metadata to the tag-key is is stored in on the S3 object."""

    SIZE_EOCD: Final[Literal[22]] = 22
    """Size of a zip file's End of Central Directory Record (empty zip)."""

    ZIPFILE_PERMISSION_MASK: ClassVar[int] = (
        stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO
    ) << 16
    """Mask to retrieve unix file permissions from the external attributes
    property of a ``zipfile.ZipInfo``.
    """

    project: _ProjectTypeVar
    """Project that is being built into a deployment package."""

    usage_type: Literal["function", "layer"]
    """How the deployment package can be used by AWS Lambda."""

    _put_object_response: Optional[PutObjectOutputTypeDef] = None

    def __init__(
        self,
        project: _ProjectTypeVar,
        usage_type: Literal["function", "layer"] = "function",
    ) -> None:
        """Instantiate class.

        The provided ``.init()`` class method should be used in place of
        direct instantiation.

        Args:
            project: Project that is being built into a deployment package.
            usage_type: How the deployment package can be used by AWS Lambda.

        """
        self.project = project
        self.usage_type = usage_type

    @cached_property
    def archive_file(self) -> Path:
        """Path to archive file.

        Because the archive file path contains runtime, it's use can cause a
        race condition or recursion error if used in some locations.
        If we removed runtime from the path we would not have a way to track
        changes to runtime which is more important than needing to be mindful
        of where this is used.

        """
        return self.project.build_directory / (
            f"{self.project.source_code.root_directory.name}."
            + ("layer." if self.usage_type == "layer" else "")
            + f"{self.runtime}.{self.project.source_code.md5_hash}.zip"
        )

    @cached_property
    def bucket(self) -> Bucket:
        """AWS S3 bucket where deployment package will be uploaded."""
        bucket = Bucket(self.project.ctx, self.project.args.bucket_name)
        if bucket.forbidden:
            raise BucketAccessDeniedError(bucket)
        if bucket.not_found:
            raise BucketNotFoundError(bucket)
        return bucket

    @cached_property
    def code_sha256(self) -> str:
        """SHA256 of the archive file.

        Returns:
            Value to pass to CloudFormation ``AWS::Lambda::Version.CodeSha256``.

        Raises:
            FileNotFoundError: Property accessed before archive file has been built.

        """
        file_hash = FileHash(hashlib.sha256())
        file_hash.add_file(self.archive_file)
        return base64.b64encode(file_hash.digest).decode()

    @cached_property
    def compatible_architectures(self) -> Optional[List[str]]:
        """List of compatible instruction set architectures."""
        return self.project.compatible_architectures

    @cached_property
    def compatible_runtimes(self) -> Optional[List[str]]:
        """List of compatible runtimes."""
        return self.project.compatible_runtimes

    @cached_property
    def exists(self) -> bool:
        """Whether the deployment package exists."""
        if self.archive_file.exists():
            return True
        return False

    @cached_property
    def gitignore_filter(
        self,
    ) -> Optional[igittigitt.IgnoreParser]:
        """Filter to use when zipping dependencies.

        This should be overridden by subclasses if a filter should be used.

        """
        return None

    @cached_property
    def license(self) -> Optional[str]:
        """Software license for the project."""
        return self.project.license

    @cached_property
    def md5_checksum(self) -> str:
        """MD5 of the archive file.

        Returns:
            Value to pass as ContentMD5 when uploading to AWS S3.

        Raises:
            FileNotFoundError: Property accessed before archive file has been built.

        """
        file_hash = FileHash(hashlib.md5())
        file_hash.add_file(self.archive_file)
        return base64.b64encode(file_hash.digest).decode()

    @cached_property
    def object_key(self) -> str:
        """Key to use when upload object to AWS S3."""
        prefix = f"awslambda/{self.usage_type}s"
        if self.project.args.object_prefix:
            prefix = (
                f"{prefix}/{self.project.args.object_prefix.lstrip('/').rstrip('/')}"
            )
        return (  # this can't contain runtime - causes a cyclic dependency
            f"{prefix}/{self.project.source_code.root_directory.name}."
            f"{self.project.source_code.md5_hash}.zip"
        )

    @cached_property
    def object_version_id(self) -> Optional[str]:
        """S3 object version ID.

        Returns:
            The ID of the current object version. This will only have a value
            if versioning is enabled on the bucket.

        """
        if (
            not self._put_object_response
            or "VersionId" not in self._put_object_response
        ):
            return None
        return self._put_object_response["VersionId"]

    @cached_property
    def runtime(self) -> str:
        """Runtime of the deployment package."""
        return self.project.runtime

    def build(self) -> Path:
        """Build the deployment package."""
        if self.exists and self.archive_file.stat().st_size > self.SIZE_EOCD:
            LOGGER.info("build skipped; %s already exists", self.archive_file.name)
            return self.archive_file

        # we need to use runtime BEFORE the build process starts to allow runtime
        # errors to be raised early.
        LOGGER.info("building %s (%s)...", self.archive_file.name, self.runtime)
        with zipfile.ZipFile(
            self.archive_file, "w", zipfile.ZIP_DEFLATED
        ) as archive_file:
            self._build_zip_dependencies(archive_file)
            self._build_zip_source_code(archive_file)
            self._build_fix_file_permissions(archive_file)

        if self.archive_file.stat().st_size <= self.SIZE_EOCD:
            raise DeploymentPackageEmptyError(self.archive_file)

        # clear cached properties so they can recalculate
        self._del_cached_property("code_sha256", "exists", "md5_checksum")
        return self.archive_file

    def _build_fix_file_permissions(self, archive_file: zipfile.ZipFile) -> None:
        """Fix file permissions of the files contained within the archive file.

        Only need to ensure that the file is executable. Permissions will be
        change to 755 or 655 if needed. The change will occur within the
        archive file only - the original file will be unchanged.

        This should be run after all files have been added to the archive file.

        Args:
            archive_file: Archive file that is currently open and ready to be
                written to.

        """
        for file_info in archive_file.filelist:
            current_perms = (
                file_info.external_attr & self.ZIPFILE_PERMISSION_MASK
            ) >> 16
            required_perm = 0o755 if current_perms & stat.S_IXUSR != 0 else 0o644
            if current_perms != required_perm:
                LOGGER.debug(
                    "fixing file permissions for %s: %o => %o",
                    file_info.filename,
                    current_perms,
                    required_perm,
                )
                file_info.external_attr = (
                    file_info.external_attr & ~self.ZIPFILE_PERMISSION_MASK
                ) | (required_perm << 16)

    def _build_zip_dependencies(
        self,
        archive_file: zipfile.ZipFile,
    ) -> None:
        """Handle installing & zipping dependencies.

        Args:
            archive_file: Archive file that is currently open and ready to be
                written to.

        """
        self.project.install_dependencies()
        for dep in self.iterate_dependency_directory():
            archive_file.write(
                dep,
                self.insert_layer_dir(
                    dep, self.project.dependency_directory
                ).relative_to(self.project.dependency_directory)
                if self.usage_type == "layer"
                else dep.relative_to(self.project.dependency_directory),
            )

    def _build_zip_source_code(self, archive_file: zipfile.ZipFile) -> None:
        """Handle zipping the project source code.

        Args:
            archive_file: Archive file that is currently open and ready to be
                written to.

        """
        for src_file in self.project.source_code:
            archive_file.write(
                src_file,
                self.insert_layer_dir(
                    src_file, self.project.source_code.root_directory
                ).relative_to(self.project.source_code.root_directory)
                if self.usage_type == "layer"
                else src_file.relative_to(self.project.source_code.root_directory),
            )

    @overload
    def build_tag_set(self, *, url_encoded: Literal[True] = ...) -> str:
        ...

    @overload
    def build_tag_set(self, *, url_encoded: Literal[False] = ...) -> Dict[str, str]:
        ...

    @overload
    def build_tag_set(self, *, url_encoded: bool = ...) -> Union[Dict[str, str], str]:
        ...

    def build_tag_set(self, *, url_encoded: bool = True) -> Union[Dict[str, str], str]:
        """Build tag set to be applied to the S3 object.

        Args:
            layer: Tag the deployment package as a Lambda Layer or not.
            url_encoded: Whether to return a dict or URL encoded query string.

        """
        optional_metadata = {
            self.META_TAGS["compatible_architectures"]: "+".join(
                self.project.compatible_architectures
            )
            if self.project.compatible_architectures
            else None,
            self.META_TAGS["compatible_runtimes"]: "+".join(
                self.project.compatible_runtimes
            )
            if self.project.compatible_runtimes
            else None,
            self.META_TAGS["license"]: self.project.license,
        }
        metadata = {
            self.META_TAGS["code_sha256"]: self.code_sha256,
            self.META_TAGS["md5_checksum"]: self.md5_checksum,
            self.META_TAGS["runtime"]: self.runtime,
            self.META_TAGS["source_code.hash"]: self.project.source_code.md5_hash,
        }
        tags = {
            **self.project.ctx.tags,
            **self.project.args.tags,
            **metadata,
            **{k: v for k, v in optional_metadata.items() if v},
        }
        if url_encoded:
            return urlencode(tags)
        return tags

    def delete(self) -> None:
        """Delete deployment package."""
        if sys.version_info < (3, 8):  # cov: ignore
            with suppress(FileNotFoundError):  # acts the same as `missing_ok=true`
                self.archive_file.unlink()  # python3.7 does not support `missing_ok`
        else:  # cov: ignore
            self.archive_file.unlink(missing_ok=True)
        LOGGER.verbose("deleted local deployment package %s", self.archive_file)
        # clear cached properties so they can recalculate
        self._del_cached_property(
            "code_sha256", "exists", "md5_checksum", "object_version_id"
        )

    @staticmethod
    def insert_layer_dir(
        file_path: Path, relative_to: Path  # pylint: disable=unused-argument
    ) -> Path:
        """Insert directory into local file path for layer archive.

        If required, this should be overridden by a subclass for language
        specific requirements.

        Args:
            file_path: Path to local file.
            relative_to: Path to a directory that the file_path will be relative
                to in the deployment package.

        """
        return file_path

    def iterate_dependency_directory(self) -> Iterator[Path]:
        """Iterate over the contents of the dependency directory.

        If ``gitignore_filter`` is set, it will be used to exclude files.

        """
        for child in self.project.dependency_directory.rglob("*"):
            if child.is_dir():
                continue  # ignore directories
            if self.gitignore_filter and self.gitignore_filter.match(child):
                continue  # ignore files that match the filter
            yield child

    def upload(self, *, build: bool = True) -> None:
        """Upload deployment package.

        Args:
            build: If true, the deployment package will be built before before
                trying to upload it. If false, it must have already been built.

        """
        if build:
            self.build()

        # we don't really need encoding - it can be NoneType so throw it away
        content_type, _content_encoding = mimetypes.guess_type(self.archive_file)

        LOGGER.info(
            "uploading deployment package %s...",
            self.bucket.format_bucket_path_uri(key=self.object_key),
        )

        self._put_object_response = self.bucket.client.put_object(
            Body=self.archive_file.read_bytes(),
            Bucket=self.project.args.bucket_name,
            ContentMD5=self.md5_checksum,
            Key=self.object_key,
            Tagging=self.build_tag_set(),
            **{"ContentType": content_type} if content_type else {},
        )
        # clear cached properties so they can recalculate
        self._del_cached_property("object_version_id")

    @classmethod
    def init(
        cls,
        project: _ProjectTypeVar,
        usage_type: Literal["function", "layer"] = "function",
    ) -> DeploymentPackage[_ProjectTypeVar]:
        """Initialize deployment package.

        This should be used in place of creating an instance of this class
        directly as it will automatically account for the S3 object already
        existing.

        Args:
            project: Project that is being built into a deployment package.
            usage_type: How the deployment package can be used by AWS Lambda.

        Returns:
            Instance of generic S3 object class if S3 object exists else
            an instance of this class.

        """
        s3_obj = DeploymentPackageS3Object(project, usage_type)
        if s3_obj.exists:
            if s3_obj.runtime == project.runtime:
                return s3_obj
            LOGGER.warning(
                "runtime of deployment package found in S3 (%s) does not match "
                "requirement (%s); deleting & recreating...",
                s3_obj.runtime,
                project.runtime,
            )
            s3_obj.delete()
        return cls(project, usage_type)


class DeploymentPackageS3Object(DeploymentPackage[_ProjectTypeVar]):
    """AWS Lambda Deployment Package.

    This should not need to be subclassed as the interactions required should
    be universal.

    Attributes:
        project: Project that is being built into a deployment package.

    """

    @cached_property
    def code_sha256(self) -> str:
        """SHA256 of the archive file.

        Returns:
            Value to pass to CloudFormation ``AWS::Lambda::Version.CodeSha256``.

        Raises:
            RequiredTagNotFound: A required tag was not found.

        """
        if self.META_TAGS["code_sha256"] not in self.object_tags:
            raise RequiredTagNotFoundError(
                self.bucket.format_bucket_path_uri(key=self.object_key),
                self.META_TAGS["code_sha256"],
            )
        return self.object_tags[self.META_TAGS["code_sha256"]]

    @cached_property
    def compatible_architectures(self) -> Optional[List[str]]:
        """List of compatible instruction set architectures."""
        if self.META_TAGS["compatible_architectures"] in self.object_tags:
            return self.object_tags[self.META_TAGS["compatible_architectures"]].split(
                "+"
            )
        return None

    @cached_property
    def compatible_runtimes(self) -> Optional[List[str]]:
        """List of compatible runtimes."""
        if self.META_TAGS["compatible_runtimes"] in self.object_tags:
            return self.object_tags[self.META_TAGS["compatible_runtimes"]].split("+")
        return None

    @cached_property
    def exists(self) -> bool:
        """Whether the S3 object exists."""
        if self.head and not self.head.get("DeleteMarker", False):
            return True
        return False

    @cached_property
    def head(self) -> Optional[HeadObjectOutputTypeDef]:
        """Response from HeadObject API call."""
        try:
            return self.bucket.client.head_object(
                Bucket=self.bucket.name, Key=self.object_key
            )
        except self.bucket.client.exceptions.ClientError as exc:
            status_code = exc.response.get("ResponseMetadata", {}).get(
                "HTTPStatusCode", 0
            )
            if status_code == 404:
                LOGGER.verbose(
                    "%s not found",
                    self.bucket.format_bucket_path_uri(key=self.object_key),
                )
                return None
            if status_code == 403:
                # we can't handle this error but, we can enhance the error message
                LOGGER.error(
                    "access denied for object %s",
                    self.bucket.format_bucket_path_uri(key=self.object_key),
                )
            raise

    @cached_property
    def license(self) -> Optional[str]:
        """Software license for the project."""
        if self.META_TAGS["license"] in self.object_tags:
            return self.object_tags[self.META_TAGS["license"]]
        return None

    @cached_property
    def md5_checksum(self) -> str:
        """MD5 of the archive file.

        Returns:
            Value to pass as ContentMD5 when uploading to AWS S3.

        Raises:
            RequiredTagNotFoundError: A required tag was not found.

        """
        if self.META_TAGS["md5_checksum"] not in self.object_tags:
            raise RequiredTagNotFoundError(
                self.bucket.format_bucket_path_uri(key=self.object_key),
                self.META_TAGS["md5_checksum"],
            )
        return self.object_tags[self.META_TAGS["md5_checksum"]]

    @cached_property
    def object_tags(self) -> Dict[str, str]:
        """S3 object tags."""
        response = self.bucket.client.get_object_tagging(
            Bucket=self.bucket.name, Key=self.object_key
        )
        if "TagSet" not in response:
            # can't be hit when using botocore.stub.Stubber as TagSet is required
            return {}  # cov: ignore
        return {t["Key"]: t["Value"] for t in response["TagSet"]}

    @cached_property
    def object_version_id(self) -> Optional[str]:
        """S3 object version ID.

        Returns:
            The ID of the current object version. This will only have a value
            if versioning is enabled on the bucket.

        """
        if not self.head or "VersionId" not in self.head:
            return None
        version_id = self.head["VersionId"]
        return version_id if version_id != "null" else None

    @cached_property
    def runtime(self) -> str:
        """Runtime of the deployment package.

        Raises:
            RequiredTagNotFoundError: A required tag was not found.

        """
        if self.META_TAGS["runtime"] not in self.object_tags:
            raise RequiredTagNotFoundError(
                self.bucket.format_bucket_path_uri(key=self.object_key),
                self.META_TAGS["runtime"],
            )
        return self.object_tags[self.META_TAGS["runtime"]]

    def build(self) -> Path:
        """Build the deployment package.

        The object should already exist. This method only exists as a "placeholder"
        to match the parent class. If the object does not already exist, and
        error is raised.

        Raises:
            S3ObjectDoesNotExistError: The S3 object does not exist.

        """
        if not self.exists:
            raise S3ObjectDoesNotExistError(self.bucket.name, self.object_key)
        LOGGER.info("build skipped; %s already exists", self.archive_file.name)
        return self.archive_file

    def delete(self) -> None:
        """Delete deployment package."""
        if self.exists:
            self.bucket.client.delete_object(
                Bucket=self.bucket.name, Key=self.object_key
            )
            LOGGER.verbose(
                "deleted deployment package S3 object %s",
                self.bucket.format_bucket_path_uri(key=self.object_key),
            )
            # clear cached properties so they can recalculate
            self._del_cached_property(
                "code_sha256",
                "exists",
                "md5_checksum",
                "object_tags",
                "object_version_id",
                "runtime",
            )

    def update_tags(self) -> None:
        """Update tags of the S3 object."""
        new_tags = self.build_tag_set(url_encoded=False)
        if new_tags == self.object_tags:
            LOGGER.debug(
                "%s tags don't need to be updated",
                self.bucket.format_bucket_path_uri(key=self.object_key),
            )
            return
        self.bucket.client.put_object_tagging(
            Bucket=self.bucket.name,
            Key=self.object_key,
            Tagging={"TagSet": [{"Key": k, "Value": v} for k, v in new_tags.items()]},
        )
        LOGGER.info("updated S3 object's tags")

    # pylint: disable=unused-argument
    def upload(self, *, build: bool = True) -> None:
        """Upload deployment package.

        The object should already exist. This method only exists as a "placeholder"
        to match the parent class. If the object does not already exist, and
        error is raised.

        Args:
            build: If true, the deployment package will be built before before
                trying to upload it. If false, it must have already been built.

        Raises:
            S3ObjectDoesNotExistError: The S3 object does not exist.

        """
        if not self.exists:
            raise S3ObjectDoesNotExistError(self.bucket.name, self.object_key)
        LOGGER.info(
            "upload skipped; %s already exists",
            self.bucket.format_bucket_path_uri(key=self.object_key),
        )
        self.update_tags()



"""CFNgin exceptions."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Union

from ..exceptions import RunwayError

if TYPE_CHECKING:
    from ..type_defs import AnyPath
    from ..variables import Variable
    from .plan import Step


class CfnginError(RunwayError):
    """Base class for custom exceptions raised by Runway."""

    message: str


class CancelExecution(CfnginError):
    """Raised when we want to cancel executing the plan."""

    message: str = "Plan canceled"


class CfnginBucketAccessDenied(CfnginError):
    """Access denied to CFNgin bucket.

    This can occur when the bucket exists in another AWS account and/or the
    credentials being used do not have adequate permissions to access the bucket.

    """

    bucket_name: str
    message: str

    def __init__(self, *, bucket_name: str) -> None:
        """Instantiate class.

        Args:
            bucket_name: Name of the CFNgin bucket.

        """
        self.bucket_name = bucket_name
        self.message = f"access denied for cfngin_bucket {bucket_name}"
        super().__init__()


class CfnginBucketNotFound(CfnginError):
    """CFNgin bucket specified or default bucket being used but it does not exist.

    This can occur when using a custom stack to deploy the CFNgin bucket but the
    custom stack does not create bucket that is expected.

    """

    bucket_name: str
    message: str

    def __init__(self, *, bucket_name: str) -> None:
        """Instantiate class.

        Args:
            bucket_name: Name of the CFNgin bucket.

        """
        self.bucket_name = bucket_name
        self.message = f"cfngin_bucket does not exist {bucket_name}"
        super().__init__()


class CfnginBucketRequired(CfnginError):
    """CFNgin bucket is required to use a feature but it not provided/disabled."""

    config_path: Optional[Path]
    message: str

    def __init__(
        self, *, config_path: Optional[AnyPath] = None, reason: Optional[str] = None
    ) -> None:
        """Instantiate class.

        Args:
            config_path: Path to the CFNgin config file.
            reason: Reason why CFNgin bucket is needed.

        """
        self.message = "cfngin_bucket is required"
        if reason:
            self.message += f"; {reason}"
        if isinstance(config_path, str):
            config_path = Path(config_path)
        if config_path:
            self.message += f" ({config_path})"
        self.config_path = config_path
        super().__init__()


class CfnginOnlyLookupError(CfnginError):
    """Attempted to use a CFNgin lookup outside of CFNgin."""

    lookup_name: str

    def __init__(self, lookup_name: str) -> None:
        """Instantiate class."""
        self.lookup_name = lookup_name
        self.message = (
            f"attempted to use CFNgin only lookup {lookup_name} outside of CFNgin"
        )
        super().__init__()


class ChangesetDidNotStabilize(CfnginError):
    """Raised when the applying a changeset fails."""

    message: str

    def __init__(self, change_set_id: str) -> None:
        """Instantiate class.

        Args:
            change_set_id: The changeset that failed.

        """
        self.id = change_set_id
        self.message = f"Changeset '{change_set_id}' did not reach a completed state."
        super().__init__()


class GraphError(CfnginError):
    """Raised when the graph is invalid (e.g. acyclic dependencies)."""

    message: str

    def __init__(self, exception: Exception, stack: str, dependency: str) -> None:
        """Instantiate class.

        Args:
            exception: The exception that was raised by the invalid
                graph.
            stack: Name of the stack causing the error.
            dependency: Name of the dependency causing the error.

        """
        self.stack = stack
        self.dependency = dependency
        self.exception = exception
        self.message = (
            f"Error detected when adding '{dependency}' "
            f"as a dependency of '{stack}': {exception}"
        )
        super().__init__()


class ImproperlyConfigured(CfnginError):
    """Raised when a component is improperly configured."""

    message: str

    def __init__(self, kls: Any, error: Exception, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            kls: The class that was improperly configured.
            error: The exception that was raised when trying to use cls.

        """
        self.message = f'Class "{kls}" is improperly configured: {error}'
        super().__init__(*args, **kwargs)


class InvalidConfig(CfnginError):
    """Provided config file is invalid."""

    errors: Union[str, List[Union[Exception, str]]]
    message: str

    def __init__(self, errors: Union[str, List[Union[Exception, str]]]) -> None:
        """Instantiate class.

        Args:
            errors: Errors or error messages that are raised to identify that a
                config is invalid.

        """
        self.errors = errors
        if isinstance(errors, list):
            self.message = "\n".join(str(e) for e in errors)
        else:
            self.message = errors
        super().__init__(errors)


class InvalidDockerizePipConfiguration(CfnginError):
    """Raised when the provided configuration for dockerized pip is invalid."""

    message: str

    def __init__(self, msg: str) -> None:
        """Instantiate class.

        Args:
            msg: The reason for the error being raised.

        """
        self.message = msg
        super().__init__()


class InvalidUserdataPlaceholder(CfnginError):
    """Raised when a placeholder name in raw_user_data is not valid.

    E.g ``${100}`` would raise this.

    """

    message: str

    def __init__(
        self, blueprint_name: str, exception_message: str, *args: Any, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            blueprint_name: Name of the blueprint with invalid userdata placeholder.
            exception_message: Message from the exception that was raised while
                parsing the userdata.

        """
        self.message = (
            f'{exception_message}. Could not parse userdata in blueprint {blueprint_name}". '
            "Make sure to escape all $ symbols with a $$."
        )
        super().__init__(*args, **kwargs)


class MissingEnvironment(CfnginError):
    """Raised when an environment lookup is used but the key doesn't exist."""

    message: str

    def __init__(self, key: str, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            key: The key that was used but doesn't exist in the environment.

        """
        self.key = key
        self.message = f"Environment missing key {key}."
        super().__init__(*args, **kwargs)


class MissingParameterException(CfnginError):
    """Raised if a required parameter with no default is missing."""

    message: str

    def __init__(self, parameters: List[str], *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            parameters: A list of the parameters that are missing.

        """
        self.parameters = parameters
        self.message = (
            f"Missing required cloudformation parameters: {', '.join(parameters)}"
        )
        super().__init__(*args, **kwargs)


class MissingVariable(CfnginError):
    """Raised when a variable with no default is not provided a value."""

    message: str

    def __init__(
        self, blueprint_name: str, variable_name: str, *args: Any, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            blueprint_name: Name of the blueprint.
            variable_name: Name of the variable missing a value.

        """
        self.message = (
            f'Variable "{variable_name}" in blueprint "{blueprint_name}" is missing'
        )
        super().__init__(*args, **kwargs)


class PipError(CfnginError):
    """Raised when pip returns a non-zero exit code."""

    message: str

    def __init__(self) -> None:
        """Instantiate class."""
        self.message = (
            "A non-zero exit code was returned when invoking "
            "pip. More information can be found in the error above."
        )
        super().__init__()


class PipenvError(CfnginError):
    """Raised when pipenv returns a non-zero exit code."""

    message: str

    def __init__(self) -> None:
        """Instantiate class."""
        self.message = (
            "A non-zero exit code was returned when invoking "
            "pipenv. Please ensure pipenv in installed and the "
            "Pipfile being used is valid. More information can be "
            "found in the error above."
        )
        super().__init__()


class PersistentGraphCannotLock(CfnginError):
    """Raised when the persistent graph in S3 cannot be locked."""

    message: str

    def __init__(self, reason: str) -> None:
        """Instantiate class."""
        self.message = f"Could not lock persistent graph; {reason}"
        super().__init__()


class PersistentGraphCannotUnlock(CfnginError):
    """Raised when the persistent graph in S3 cannot be unlocked."""

    message: str

    def __init__(self, reason: Union[Exception, str]) -> None:
        """Instantiate class."""
        self.message = f"Could not unlock persistent graph; {reason}"
        super().__init__()


class PersistentGraphLocked(CfnginError):
    """Raised when the persistent graph in S3 is lock.

    The action being executed requires it to be unlocked before attempted.

    """

    message: str

    def __init__(
        self, *, message: Optional[str] = None, reason: Optional[str] = None
    ) -> None:
        """Instantiate class."""
        if message:
            self.message = message
        else:
            reason = (
                reason
                or "This action requires the graph to be unlocked to be executed."
            )
            self.message = f"Persistent graph is locked. {reason}"
        super().__init__()


class PersistentGraphLockCodeMismatch(CfnginError):
    """Raised when the provided persistent graph lock code does not match.

    The code used to unlock the persistent graph must match the s3 object lock
    code.

    """

    message: str

    def __init__(self, provided_code: str, s3_code: Optional[str]) -> None:
        """Instantiate class."""
        self.message = (
            f"The provided lock code '{provided_code}' does not match the S3 "
            f"object lock code '{s3_code}'"
        )
        super().__init__()


class PersistentGraphUnlocked(CfnginError):
    """Raised when the persistent graph in S3 is unlock.

    The action being executed requires it to be locked before attempted.

    """

    message: str

    def __init__(
        self, message: Optional[str] = None, reason: Optional[str] = None
    ) -> None:
        """Instantiate class."""
        if message:
            self.message = message
        else:
            reason = (
                reason or "This action requires the graph to be locked to be executed."
            )
            self.message = f"Persistent graph is unlocked. {reason}"
        super().__init__()


class PlanFailed(CfnginError):
    """Raised if any step of a plan fails."""

    message: str

    def __init__(self, failed_steps: List[Step], *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            failed_steps: The steps that failed.

        """
        self.failed_steps = failed_steps

        step_names = ", ".join(step.name for step in failed_steps)
        self.message = f"The following steps failed: {step_names}"

        super().__init__(*args, **kwargs)


class StackDidNotChange(CfnginError):
    """Raised when there are no changes to be made by the provider."""

    message: str = "Stack did not change"


class StackDoesNotExist(CfnginError):
    """Raised when a stack does not exist in AWS."""

    message: str

    def __init__(self, stack_name: str, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            stack_name: Name of the stack that does not exist.

        """
        self.message = (
            f'Stack: "{stack_name}" does not exist in outputs or the lookup is '
            "not available in this CFNgin run"
        )
        super().__init__(*args, **kwargs)


class StackUpdateBadStatus(CfnginError):
    """Raised if the state of a stack can't be handled."""

    message: str

    def __init__(
        self, stack_name: str, stack_status: str, reason: str, *args: Any, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            stack_name: Name of the stack.
            stack_status: The stack's status.
            reason: The reason for the current status.

        """
        self.stack_name = stack_name
        self.stack_status = stack_status

        self.message = (
            f'Stack: "{stack_name}" cannot be updated nor re-created from state '
            f"{stack_status}: {reason}"
        )
        super().__init__(*args, **kwargs)


class StackFailed(CfnginError):
    """Raised when a stack action fails.

    Primarily used with hooks that act on stacks.

    """

    message: str

    def __init__(self, stack_name: str, status_reason: Optional[str] = None) -> None:
        """Instantiate class.

        Args:
            stack_name: Name of the stack.
            status_reason: The reason for the current status.

        """
        self.stack_name = stack_name
        self.status_reason = status_reason

        self.message = f'Stack "{stack_name}" failed'
        if status_reason:
            self.message += f' with reason "{status_reason}"'
        super().__init__()


class UnableToExecuteChangeSet(CfnginError):
    """Raised if changeset execution status is not ``AVAILABLE``."""

    message: str

    def __init__(
        self, stack_name: str, change_set_id: str, execution_status: str
    ) -> None:
        """Instantiate class.

        Args:
            stack_name: Name of the stack.
            change_set_id: The changeset that failed.
            execution_status: The value of the changeset's ``ExecutionStatus``
                attribute.

        """
        self.stack_name = stack_name
        self.id = change_set_id
        self.execution_status = execution_status

        self.message = (
            f"Changeset '{change_set_id}' on stack '{stack_name}' had bad "
            f"execution status: {execution_status}"
        )

        super().__init__()


class UnhandledChangeSetStatus(CfnginError):
    """Raised when creating a changeset failed for an unhandled reason.

    Handled failure reasons include: no changes

    """

    message: str

    def __init__(
        self, stack_name: str, change_set_id: str, status: str, status_reason: str
    ) -> None:
        """Instantiate class.

        Args:
            stack_name: Name of the stack.
            change_set_id: The changeset that failed.
            status: The state that could not be handled.
            status_reason: Cause of the current state.

        """
        self.stack_name = stack_name
        self.id = change_set_id
        self.status = status
        self.status_reason = status_reason
        self.message = (
            f"Changeset '{change_set_id}' on stack '{stack_name}' returned an unhandled status "
            f"'{status}: {status_reason}'."
        )

        super().__init__()


class UnresolvedBlueprintVariable(CfnginError):
    """Raised when trying to use a variable before it has been resolved."""

    message: str

    def __init__(
        self, blueprint_name: str, variable: Variable, *args: Any, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            blueprint_name: Name of the blueprint that tried to use
                the unresolved variables.
            variable: The unresolved variable.

        """
        self.message = (
            f'Variable "{variable.name}" in blueprint "{blueprint_name}" '
            "hasn't been resolved"
        )
        super().__init__(*args, **kwargs)


class UnresolvedBlueprintVariables(CfnginError):
    """Raised when trying to use variables before they has been resolved."""

    message: str

    def __init__(self, blueprint_name: str, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            blueprint_name: Name of the blueprint that tried to use the unresolved
                variables.

        """
        self.message = f"Blueprint: \"{blueprint_name}\" hasn't resolved it's variables"
        super().__init__(*args, **kwargs)


class ValidatorError(CfnginError):
    """Used for errors raised by custom validators of blueprint variables."""

    message: str

    def __init__(
        self,
        variable: str,
        validator: str,
        value: str,
        exception: Optional[Exception] = None,
    ) -> None:
        """Instantiate class.

        Args:
            variable: The variable that failed validation.
            validator: The validator that was not passed.
            value: The value of the variable that did not pass the validator.
            exception: The exception raised by the validator.

        """
        self.variable = variable
        self.validator = validator
        self.value = value
        self.exception = exception
        self.message = (
            f"Validator '{self.validator}' failed for variable '{self.variable}' "
            f"with value '{self.value}'"
        )

        if self.exception:
            self.message += (
                f": {self.exception.__class__.__name__}: {str(self.exception)}"
            )
        super().__init__()

    def __str__(self):
        """Return the exception's message when converting to a string."""
        return self.message


class VariableTypeRequired(CfnginError):
    """Raised when a variable defined in a blueprint is missing a type."""

    message: str

    def __init__(
        self, blueprint_name: str, variable_name: str, *args: Any, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            blueprint_name: Name of the blueprint.
            variable_name: Name of the variable missing a type.

        """
        self.message = (
            f'Variable "{variable_name}" in blueprint "{blueprint_name}" '
            "does not have a type"
        )
        super().__init__(*args, **kwargs)



"""Docker hook_data object."""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, overload

from docker import DockerClient

from ....compat import cached_property
from ....utils import MutableMap

if TYPE_CHECKING:
    from ....context import CfnginContext
    from .data_models import DockerImage


class DockerHookData(MutableMap):
    """Docker hook_data object."""

    image: Optional["DockerImage"] = None

    @cached_property
    def client(self) -> DockerClient:
        """Docker client."""
        return DockerClient.from_env()

    @overload
    def update_context(self, context: CfnginContext = ...) -> DockerHookData:  # noqa
        ...

    @overload
    def update_context(self, context: None = ...) -> None:  # noqa
        ...

    def update_context(
        self, context: Optional[CfnginContext] = None
    ) -> Optional[DockerHookData]:
        """Update context object with new the current DockerHookData."""
        if not context:
            return None
        context.hook_data["docker"] = self
        return self

    @classmethod
    def from_cfngin_context(cls, context: CfnginContext) -> DockerHookData:
        """Get existing object or create a new one."""
        if "docker" in context.hook_data:
            found = context.hook_data["docker"]
            if isinstance(found, cls):
                return found
        new_obj = cls()
        context.hook_data["docker"] = new_obj
        return new_obj



"""File lookup."""
# pylint: disable=arguments-differ,no-self-argument
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

import base64
import collections.abc
import json
import re
from typing import Any, Callable, Dict, List, Mapping, Sequence, Tuple, Union, overload

import yaml
from pydantic import validator
from troposphere import Base64, GenericHelperFn
from typing_extensions import Final, Literal

from ....lookups.handlers.base import LookupHandler
from ....utils import BaseModel
from ...utils import read_value_from_path

_PARAMETER_PATTERN = re.compile(r"{{([::|\w]+)}}")

ParameterizedObjectTypeDef = Union[str, Mapping[str, Any], Sequence[Any], Any]
ParameterizedObjectReturnTypeDef = Union[
    Dict[str, "ParameterizedObjectReturnTypeDef"],
    GenericHelperFn,
    List["ParameterizedObjectReturnTypeDef"],
]


class ArgsDataModel(BaseModel):
    """Arguments data model."""

    codec: str
    """Codec that will be used to parse and/or manipulate the data."""

    @validator("codec", allow_reuse=True)
    def _validate_supported_codec(cls, v: str) -> str:
        """Validate that the selected codec is supported."""
        if v in CODECS:
            return v
        raise ValueError(f"Codec '{v}' must be one of: {', '.join(CODECS)}")


class FileLookup(LookupHandler):
    """File lookup."""

    TYPE_NAME: Final[Literal["file"]] = "file"
    """Name that the Lookup is registered as."""

    @classmethod
    def parse(cls, value: str) -> Tuple[str, Dict[str, str]]:
        """Parse the value passed to the lookup.

        This overrides the default parsing to account for special requirements.

        Args:
            value: The raw value passed to a lookup.

        Returns:
            The lookup query and a dict of arguments

        Raises:
            ValueError: The value provided does not match the expected regex.

        """
        args: Dict[str, str] = {}
        try:
            args["codec"], data_or_path = value.split(":", 1)
        except ValueError:
            raise ValueError(
                f"Query '{value}' doesn't match regex: "
                rf"^(?P<codec>[{'|'.join(CODECS)}]:.+$)"
            ) from None
        return read_value_from_path(data_or_path), args

    @classmethod
    def handle(cls, value: str, **_: Any) -> Any:
        """Translate a filename into the file contents."""
        data, raw_args = cls.parse(value)
        args = ArgsDataModel.parse_obj(raw_args)
        return CODECS[args.codec](data)


def _parameterize_string(raw: str) -> GenericHelperFn:
    """Substitute placeholders in a string using CloudFormation references.

    Args:
        raw: String to be processed. Byte strings are not supported; decode them
        before passing them to this function.

    Returns:
        An expression with placeholders from the input replaced, suitable to be
        passed to Troposphere to be included in CloudFormation template.
        This will be the input string without modification if no substitutions
        are found, and a composition of CloudFormation calls otherwise.

    """
    parts: List[Any] = []
    s_index = 0

    for match in _PARAMETER_PATTERN.finditer(raw):
        parts.append(raw[s_index : match.start()])
        parts.append({"Ref": match.group(1)})
        s_index = match.end()

    if not parts:
        return GenericHelperFn(raw)

    parts.append(raw[s_index:])
    return GenericHelperFn({"Fn::Join": ["", parts]})


@overload
def parameterized_codec(raw: str, b64: Literal[False] = ...) -> GenericHelperFn:
    ...


@overload
def parameterized_codec(raw: str, b64: Literal[True] = ...) -> Base64:
    ...


def parameterized_codec(raw: str, b64: bool = False) -> Any:
    """Parameterize a string, possibly encoding it as Base64 afterwards.

    Args:
        raw: String to be processed. Byte strings will be interpreted as UTF-8.
        b64: Whether to wrap the output in a Base64 CloudFormation call.

    Returns:
        :class:`troposphere.AWSHelperFn`: Output to be included in a
        CloudFormation template.

    """
    result = _parameterize_string(raw)
    # Note, since we want a raw JSON object (not a string) output in the
    # template, we wrap the result in GenericHelperFn (not needed if we're
    # using Base64)
    return Base64(result.data) if b64 else result


@overload
def _parameterize_obj(obj: Union[bytes, str]) -> GenericHelperFn:
    ...


@overload
def _parameterize_obj(obj: Mapping[str, Any]) -> ParameterizedObjectReturnTypeDef:
    ...


@overload
def _parameterize_obj(obj: List[Any]) -> ParameterizedObjectReturnTypeDef:
    ...


def _parameterize_obj(
    obj: ParameterizedObjectTypeDef,
) -> ParameterizedObjectReturnTypeDef:
    """Recursively parameterize all strings contained in an object.

    Parametrizes all values of a Mapping, all items of a Sequence, an
    unicode string, or pass other objects through unmodified.

    Args:
        obj: Data to parameterize.

    Return:
        A parameterized object to be included in a CloudFormation template.
        Mappings are converted to `dict`, Sequences are converted to  `list`,
        and strings possibly replaced by compositions of function calls.

    """
    if isinstance(obj, str):
        return _parameterize_string(obj)
    if isinstance(obj, collections.abc.Mapping):
        return {key: _parameterize_obj(value) for key, value in obj.items()}
    if isinstance(obj, collections.abc.Sequence):
        return [_parameterize_obj(item) for item in obj]
    return obj


def yaml_codec(raw: str, parameterized: bool = False) -> Any:
    """YAML codec."""
    data = yaml.load(raw, Loader=yaml.SafeLoader)
    return _parameterize_obj(data) if parameterized else data


def json_codec(raw: str, parameterized: bool = False) -> Any:
    """JSON codec."""
    data = json.loads(raw)
    return _parameterize_obj(data) if parameterized else data


CODECS: Dict[str, Callable[[str], Any]] = {
    "base64": lambda x: base64.b64encode(x.encode("utf8")).decode("utf-8"),
    "json": lambda x: json_codec(x, parameterized=False),
    "json-parameterized": lambda x: json_codec(x, parameterized=True),
    "parameterized": lambda x: parameterized_codec(x, False),
    "parameterized-b64": lambda x: parameterized_codec(x, True),
    "plain": lambda x: x,
    "yaml": lambda x: yaml_codec(x, parameterized=False),
    "yaml-parameterized": lambda x: yaml_codec(x, parameterized=True),
}



"""Classes for interacting with dependency managers using subprocesses."""
from . import base_classes
from ._pip import Pip, PipInstallFailedError
from ._pipenv import Pipenv, PipenvExportFailedError, PipenvNotFoundError
from ._poetry import Poetry, PoetryExportFailedError, PoetryNotFoundError

__all__ = [
    "Pip",
    "PipInstallFailedError",
    "Pipenv",
    "PipenvExportFailedError",
    "PipenvNotFoundError",
    "Poetry",
    "PoetryExportFailedError",
    "PoetryNotFoundError",
    "base_classes",
]



"""Runway Static Site Module options component classes."""
from __future__ import annotations

from ...base import ModuleOptions
from .models import RunwayStaticSiteModuleOptionsDataModel


class StaticSiteOptions(ModuleOptions):
    """Static site options.

    Attributes:
        build_output: Directory where build output is placed. Defaults to current
            working directory.
        build_steps: List of commands to run to build the static site.
        data: Options parsed into a data model.
        extra_files: List of files that should be uploaded to S3 after the build.
            Used to dynamically create or select file.
        pre_build_steps: Commands to be run prior to the build process.
        source_hashing: Overrides for source hash calculation and tracking.

    """

    def __init__(self, data: RunwayStaticSiteModuleOptionsDataModel) -> None:
        """Instantiate class."""
        self.build_output = data.build_output
        self.build_steps = data.build_steps
        self.data = data
        self.extra_files = data.extra_files
        self.pre_build_steps = data.pre_build_steps
        self.source_hashing = data.source_hashing

    @classmethod
    def parse_obj(cls, obj: object) -> StaticSiteOptions:
        """Parse options definition and return an options object.

        Args:
            obj: Object to parse.

        """
        return cls(data=RunwayStaticSiteModuleOptionsDataModel.parse_obj(obj))



"""Abstract parent class for a 'Source' type object.

Allows us to specify specific remote sourced resources for out application
(Git, S3, ect.)

"""
import logging
from pathlib import Path
from typing import Any, Union

LOGGER = logging.getLogger(__name__)


class Source:
    """Abstract parent class for a 'Source' type object.

    The Source parent class allows us to specify remote resources
    for our application via services such as Git or S3.  A cache
    directory, as part of object's configuration, is automatically
    created by default: ``./.runway/cache``.
    This folder can be overridden by specifying the ``cache_dir`` property
    in the configuration passed.

    Every Source type object is expected to have a ``fetch`` method which
    will return the folder path at where the module requested resides.

    """

    cache_dir: Path

    def __init__(self, *, cache_dir: Union[Path, str], **_: Any):
        """Source.

        Args:
            cache_dir: The directory where the given remote resource should be
                cached.

        """
        self.cache_dir = cache_dir if isinstance(cache_dir, Path) else Path(cache_dir)

        self.__create_cache_directory()

    def fetch(self) -> Path:
        """Retrieve remote source. To be implemented in each subclass."""
        raise NotImplementedError

    def __create_cache_directory(self) -> None:
        """If no cache directory exists for the remote runway modules, create one."""
        self.cache_dir.mkdir(exist_ok=True, parents=True)

    @staticmethod
    def sanitize_directory_path(uri: str) -> str:
        """Sanitize a Source directory path string.

        Arguments:
            uri: The uniform resource identifier when targeting a remote resource.

        """
        for i in ["@", "/", ":"]:
            uri = uri.replace(i, "_")
        return uri
